# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["BaseCsc"]

import argparse
import asyncio
import enum
import sys
import warnings

from . import base
from . import dds_utils
from .sal_enums import State
from .controller import Controller
from .csc_utils import make_state_transition_dict

HEARTBEAT_INTERVAL = 1  # seconds


class BaseCsc(Controller):
    """Base class for a Commandable SAL Component (CSC)

    To implement a CSC in Python define a subclass of this class.

    Parameters
    ----------
    name : `str`
        Name of SAL component.
    index : `int` or `None`
        SAL component index, or 0 or None if the component is not indexed.
    initial_state : `State` or `int`, optional
        Initial state for this CSC.
        Typically `State.STANDBY` (or `State.OFFLINE` for an
        externally commandable CSC) but can also be
        `State.DISABLED`, or `State.ENABLED`,
        in which case you may also want to specify
        ``settings_to_apply`` for a configurable CSC.
    settings_to_apply : `str`, optional
        Settings to apply if ``initial_state`` is `State.DISABLED`
        or `State.ENABLED`. Ignored if the CSC is not configurable.
    simulation_mode : `int`, optional
        Simulation mode. The default is 0: do not simulate.

    Raises
    ------
    ValueError
        If ``initial_state`` is invalid.
    salobj.ExpectedError
        If ``simulation_mode`` is invalid.
        Note: you will only see this error if you await `start_task`.

    Attributes
    ----------
    default_initial_state : `State`
        A *class* attribute.
        Default initial state. Leave at `State.STANDBY` for most CSCs;
        set to `State.OFFLINE` for externally commandable CSCs.
    enable_cmdline_state : `State`
        A *class* attribute.
        If `True` then add ``--state`` and (if a subclass of `ConfigurableCsc`)
        ``--settings`` command-line arguments.
        The default is `False` because CSCs should start in
        ``default_initial_state`` unless we have good reason to do otherwise.
    require_settings : `bool`
        A *class* attribute.
        Controls whether the ``--settings`` command-line argument
        is required if ``--state`` is specified.
        Ignored unless ``enable_cmdline_state`` is True
        and the CSC is a subclass of `ConfigurableCsc`.
    simulation_help : `str`
        A *class* attribute.
        Help for the --simulate command, or None for the default help
        (which is fine for the usual case of 0/1).
        Ignored if simulation is not supported.
    valid_simulation_modes : `list` [`int`] or `None`
        A *class* attribute. A value of `None` is deprecated
        (though it is the default, for backwards compatibility;
        once `None` is prohibited there will be no default).
        If not `None` has the following effect:

        * ``simulation_mode`` will be checked in the constructor
        * `implement_simulation_mode` will be a no-op.
        * The `amain` command parser will have a ``--simulate`` argument.
          The default value will be 0 if that is a valid simulation mode
          (and it certainly should be).
          Otherwise the default value will be the first entry.
    version : `str`
        A *class* attribute. Used to set the ``cscVersion`` attribute of the
        ``softwareVersions`` event. You should almost always set this to
        your package's ``__version__`` attribute.
    heartbeat_interval : `float`
        Interval between heartbeat events, in seconds;

    Notes
    -----
    **Constructor**

    The constructor does the following, beyond the parent class constructor:

    * For each command defined for the component, find a ``do_<name>`` method
      (which must be present) and add it as a callback to the
      ``cmd_<name>`` attribute.
    * The base class provides asynchronous ``do_<name>`` methods for the
      standard CSC commands. The default implementation is as follows
      (if any step fails then the remaining steps are not performed):

        * Check for validity of the requested state change;
          if the change is valid then:
        * Call ``begin_<name>``. This is a no-op in the base class,
          and is available for the subclass to override. If this fails,
          then log an error and acknowledge the command as failed.
        * Change self.summary_state.
        * Call ``end_<name>``. Again, this is a no-op in the base class,
          and is available for the subclass to override.
          If this fails then revert ``self.summary_state``, log an error,
          and acknowledge the command as failed.
        * Call `handle_summary_state` and `report_summary_state`
          to handle report the new summary state.
          If this fails then leave the summary state updated
          (since the new value *may* have been reported),
          but acknowledge the command as failed.
        * Acknowledge the command as complete.

    * Set the summary state.
    * Run `start` asynchronously.
    """

    # See Attributes in the doc string.
    default_initial_state = State.STANDBY
    enable_cmdline_state = False
    require_settings = False
    valid_simulation_modes = None
    simulation_help = None

    def __init__(
        self,
        name,
        index=None,
        initial_state=State.STANDBY,
        settings_to_apply="",
        simulation_mode=0,
    ):
        # Check for consistent require_settings and enable_cmdline_state.
        # Do that here, instead of in make_from_cmd_line, to  make sure
        # the developer sees the problem.
        if self.require_settings and not self.enable_cmdline_state:
            raise RuntimeError(
                "Class variable require_settings=True "
                "requires class variable enable_cmdline_state=True"
            )

        # cast initial_state from an int or State to a State,
        # and reject invalid int values with ValueError
        initial_state = State(initial_state)
        if initial_state == State.FAULT:
            raise ValueError("initial_state cannot be FAULT")

        if self.valid_simulation_modes is None:
            warnings.warn(
                "valid_simulation_modes=None is deprecated", DeprecationWarning
            )
        else:
            if simulation_mode not in self.valid_simulation_modes:
                raise ValueError(
                    f"simulation_mode={simulation_mode} "
                    f"not in valid_simulation_modes={self.valid_simulation_modes}"
                )

        self._settings_to_apply = settings_to_apply
        self._summary_state = State(self.default_initial_state)
        self._initial_state = initial_state
        self._faulting = False
        self._heartbeat_task = base.make_done_future()
        # Interval between heartbeat events (sec)
        self.heartbeat_interval = HEARTBEAT_INTERVAL

        if not hasattr(self, "version"):
            warnings.warn(
                "Please set class attribute `version`. It is needed to set "
                "the `cscVersion` field of the `softwareVersions` event.",
                DeprecationWarning,
            )

        # Postpone assigning command callbacks until `start` is done (but call
        # assert_do_methods_present to fail early if there is a problem).
        super().__init__(name=name, index=index, do_callbacks=True)

        # Set evt_simulationMode, now that it is available.
        self.evt_simulationMode.set(mode=int(simulation_mode))

        def format_version(version):
            return "?" if version is None else version

        self.evt_softwareVersions.set(
            salVersion=format_version(self.salinfo.metadata.sal_version),
            xmlVersion=format_version(self.salinfo.metadata.xml_version),
            openSpliceVersion=dds_utils.get_dds_version(),
            cscVersion=getattr(self, "version", "?"),
        )

    async def start(self):
        """Finish constructing the CSC.

        * Call `set_simulation_mode`. If this fails, set ``self.start_task``
          to the exception, call `stop`, making the CSC unusable, and return.
        * Call `handle_summary_state` and `report_summary_state`.
        * Set ``self.start_task`` done.
        """
        await super().start()
        self._heartbeat_task.cancel()  # Paranoia
        self._heartbeat_task = asyncio.create_task(self._heartbeat_loop())
        await self.set_simulation_mode(self.simulation_mode)

        # Handle initial state, then transition to the desired state.
        # If this fails then log the exception and continue,
        # in hopes that the CSC can still be used. For instance
        # if the settings were invalid, the user can specify others.
        await self.handle_summary_state()
        self.report_summary_state()
        command = None
        if self._initial_state != self.default_initial_state:
            state_transition_dict = make_state_transition_dict()
            # List of commands to transition from
            # current state to initial_state.
            # Use the current state instead of default_initial_state
            # because of Hexapod and Rotator, which do not know their
            # current state until they connect to the low-level controller.
            command_state_list = state_transition_dict[
                (self.summary_state, self._initial_state)
            ]
            state_transition_commands = [item[0] for item in command_state_list]
            try:
                for command in state_transition_commands:
                    method = getattr(self, f"do_{command}")
                    data = getattr(self, f"cmd_{command}").DataType()
                    if command == "start":
                        data.settingsToApply = self._settings_to_apply
                    self.log.info(f"Executing {command} command during startup")
                    await method(data)
            except Exception:
                self.log.exception(
                    f"Failed in start on state transition command {command}; continuing."
                )

        self.evt_softwareVersions.put()

    async def close_tasks(self):
        """Shut down pending tasks. Called by `close`."""
        self._heartbeat_task.cancel()

    @classmethod
    def make_from_cmd_line(cls, index, **kwargs):
        """Construct a CSC from command line arguments.

        Parameters
        ----------
        index : `int`, `enum.IntEnum`, `True`, or `None`
            If the CSC is indexed, do one of the following:

            * Specify `True` to make ``index`` a required
              command-line argument that accepts any nonzero index.
            * Specify an `enum.IntEnum` *class* to make ``index`` a required
              command-line argument that only accepts the enum values.
            * Specify a non-zero integer to use that index.
              This is rare; if the CSC is indexed then the user
              should usually be allowed to specify the index.

            If the CSC is not indexed, specify `None` or 0.
        **kwargs : `dict`, optional
            Additional keyword arguments for your CSC's constructor.

        Returns
        -------
        csc : ``cls``
            The CSC.

        Notes
        -----
        To add additional command-line arguments, override `add_arguments`
        and `add_kwargs_from_args`.
        """
        parser = argparse.ArgumentParser(f"Run {cls.__name__}")
        if index is True:
            parser.add_argument("index", type=int, help="SAL index.")
        elif isinstance(index, type) and issubclass(index, enum.IntEnum):
            # The isinstance check just above prevents errors
            # when index is an int or bool.
            choices = [int(item.value) for item in index]
            names_str = ", ".join(
                f"{item.value}: {item.name.lower()}" for item in index
            )
            help_text = f"SAL index, one of: {names_str}"
            parser.add_argument("index", type=int, help=help_text, choices=choices)
        parser.add_argument(
            "--loglevel",
            type=int,
            help="log level: error=40, warning=30, info=20, debug=10",
        )
        if cls.enable_cmdline_state:
            parser.add_argument(
                "--state",
                choices=["offline", "standby", "disabled", "enabled"],
                dest="initial_state",
                help="initial state",
            )
        add_simulate_arg = (
            cls.valid_simulation_modes is not None
            and len(cls.valid_simulation_modes) > 1
        )
        if add_simulate_arg:
            if 0 in cls.valid_simulation_modes:
                default_simulation_mode = 0
            else:
                default_simulation_mode = cls.valid_simulation_modes[0]
            if default_simulation_mode == 0 and len(cls.valid_simulation_modes) == 2:
                # There are only two simulation modes, one of which is 0.
                # Make --simulate a flag that takes no value and stores
                # the other value, if specified.
                simulation_help = (
                    "Run in simulation mode?"
                    if cls.simulation_help is None
                    else cls.simulation_help
                )
                nonzero_value = (set(cls.valid_simulation_modes) - set([0])).pop()
                parser.add_argument(
                    "--simulate",
                    help=simulation_help,
                    default=0,
                    action="store_const",
                    const=nonzero_value,
                )
            else:
                # There are more than 2 simulation modes or none of them is 0.
                # Make --simulate an argument that requires a value.
                simulation_help = (
                    "Simulation mode"
                    if cls.simulation_help is None
                    else cls.simulation_help
                )
                parser.add_argument(
                    "--simulate",
                    type=int,
                    help=simulation_help,
                    default=default_simulation_mode,
                    choices=cls.valid_simulation_modes,
                )

        try:
            version = getattr(cls, "version", None)
            if version is not None:
                parser.add_argument("--version", action="version", version=version)
        except ImportError:
            warnings.warn(
                "No --version command-line argument because __version__ is unavailable.",
                RuntimeWarning,
            )
        cls.add_arguments(parser)

        args = parser.parse_args()
        if hasattr(args, "index"):
            kwargs["index"] = args.index
        elif not index:
            pass
        else:
            kwargs["index"] = int(index)
        if add_simulate_arg:
            kwargs["simulation_mode"] = args.simulate
        state_arg = getattr(args, "initial_state", None)
        if state_arg is not None:
            initial_state_name = state_arg.upper()
            initial_state = getattr(State, initial_state_name)
            kwargs["initial_state"] = initial_state
        if hasattr(args, "settings_to_apply"):
            # A configurable CSC with constructor arg ``settings_to_apply``
            if args.settings_to_apply is not None:
                kwargs["settings_to_apply"] = args.settings_to_apply
            elif cls.require_settings and args.initial_state in ("disabled", "enabled"):
                parser.error(
                    "You must specify --settings if you specify --state disabled or enabled"
                )
        cls.add_kwargs_from_args(args=args, kwargs=kwargs)

        csc = cls(**kwargs)
        if args.loglevel is not None:
            csc.log.setLevel(args.loglevel)
        return csc

    @classmethod
    async def amain(cls, index, **kwargs):
        """Make a CSC from command-line arguments and run it.

        Parameters
        ----------
        index : `int`, `enum.IntEnum`, `True`, or `None`
            If the CSC is indexed, do one of the following:

            * Specify `True` to make ``index`` a required
              command-line argument that accepts any nonzero index.
            * Specify an `enum.IntEnum` *class* to make ``index`` a required
              command-line argument that only accepts the enum values.
            * Specify a non-zero integer to use that index.
              This is rare; if the CSC is indexed then the user
              should usually be allowed to specify the index.

            If the CSC is not indexed, specify `None` or 0.
        **kwargs : `dict`, optional
            Additional keyword arguments for your CSC's constructor.
        """
        csc = cls.make_from_cmd_line(index=index, **kwargs)
        await csc.done_task

    @classmethod
    def add_arguments(cls, parser):
        """Add arguments to the parser created by `make_from_cmd_line`.

        Parameters
        ----------
        parser : `argparse.ArgumentParser`
            The argument parser.

        Notes
        -----
        If you override this method then you should almost certainly override
        `add_kwargs_from_args` as well.
        """
        pass

    @classmethod
    def add_kwargs_from_args(cls, args, kwargs):
        """Add constructor keyword arguments based on parsed arguments.

        Parameters
        ----------
        args : `argparse.namespace`
            Parsed command.
        kwargs : `dict`
            Keyword argument dict for the constructor.
            Update this based on ``args``.
            The index argument will already be present if relevant.

        Notes
        -----
        If you override this method then you should almost certainly override
        `add_arguments` as well.
        """
        pass

    async def do_disable(self, data):
        """Transition from `State.ENABLED` to `State.DISABLED`.

        Parameters
        ----------
        data : ``cmd_disable.DataType``
            Command data
        """
        await self._do_change_state(data, "disable", [State.ENABLED], State.DISABLED)

    async def do_enable(self, data):
        """Transition from `State.DISABLED` to `State.ENABLED`.

        Parameters
        ----------
        data : ``cmd_enable.DataType``
            Command data
        """
        await self._do_change_state(data, "enable", [State.DISABLED], State.ENABLED)

    async def do_exitControl(self, data):
        """Transition from `State.STANDBY` to `State.OFFLINE` and quit.

        Parameters
        ----------
        data : ``cmd_exitControl.DataType``
            Command data
        """
        await self._do_change_state(data, "exitControl", [State.STANDBY], State.OFFLINE)

        asyncio.create_task(self.close())

    async def do_standby(self, data):
        """Transition from `State.DISABLED` or `State.FAULT` to
        `State.STANDBY`.

        Parameters
        ----------
        data : ``cmd_standby.DataType``
            Command data
        """
        await self._do_change_state(
            data, "standby", [State.DISABLED, State.FAULT], State.STANDBY
        )

    async def do_start(self, data):
        """Transition from `State.STANDBY` to `State.DISABLED`.

        Parameters
        ----------
        data : ``cmd_start.DataType``
            Command data
        """
        await self._do_change_state(data, "start", [State.STANDBY], State.DISABLED)

    @property
    def simulation_mode(self):
        """Get the current simulation mode.

        0 means normal operation (no simulation).

        Raises
        ------
        ExpectedError
            If the new simulation mode is not a supported value.

        """
        return self.evt_simulationMode.data.mode

    async def set_simulation_mode(self, simulation_mode):
        """Set the simulation mode.

        Await implement_simulation_mode, update the simulation mode
        property and report the new value.

        Parameters
        ----------
        simulation_mode : `int`
            Requested simulation mode; 0 for normal operation.
        """
        await self.implement_simulation_mode(simulation_mode)

        self.evt_simulationMode.set_put(mode=simulation_mode, force_output=True)

    async def implement_simulation_mode(self, simulation_mode):
        """Implement going into or out of simulation mode.

        Deprecated. See :ref:`simulation mode<lsst.ts.salobj-simulation_mode>`
        for details.

        Parameters
        ----------
        simulation_mode : `int`
            Requested simulation mode; 0 for normal operation.

        Raises
        ------
        ExpectedError
            If ``simulation_mode`` is not a supported value.
        """
        if self.valid_simulation_modes is not None:
            return

        # Handle the deprecated case.
        if simulation_mode != 0:
            raise base.ExpectedError(
                f"This CSC does not support simulation; simulation_mode={simulation_mode} but must be 0"
            )

    async def begin_disable(self, data):
        """Begin do_disable; called before state changes.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def begin_enable(self, data):
        """Begin do_enable; called before state changes.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def begin_exitControl(self, data):
        """Begin do_exitControl; called before state changes.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def begin_standby(self, data):
        """Begin do_standby; called before the state changes.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def begin_start(self, data):
        """Begin do_start; called before state changes.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def end_disable(self, data):
        """End do_disable; called after state changes
        but before command acknowledged.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def end_enable(self, data):
        """End do_enable; called after state changes
        but before command acknowledged.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def end_exitControl(self, data):
        """End do_exitControl; called after state changes
        but before command acknowledged.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def end_standby(self, data):
        """End do_standby; called after state changes
        but before command acknowledged.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    async def end_start(self, data):
        """End do_start; called after state changes
        but before command acknowledged.

        Parameters
        ----------
        data : `DataType`
            Command data
        """
        pass

    def fault(self, code, report, traceback=""):
        """Enter the fault state and output the ``errorCode`` event.

        Parameters
        ----------
        code : `int`
            Error code for the ``errorCode`` event.
            If `None` then ``errorCode`` is not output and you should
            output it yourself. Specifying `None` is deprecated;
            please always specify an integer error code.
        report : `str`
            Description of the error.
        traceback : `str`, optional
            Description of the traceback, if any.
        """
        if self._faulting:
            return

        try:
            self._faulting = True
            self._summary_state = State.FAULT
            try:
                self.evt_errorCode.set_put(
                    errorCode=code,
                    errorReport=report,
                    traceback=traceback,
                    force_output=True,
                )
            except Exception:
                self.log.exception(
                    f"Failed to output errorCode: code={code!r}; report={report!r}"
                )
            try:
                self.report_summary_state()
            except Exception:
                self.log.exception(
                    "report_summary_state failed while going to FAULT; "
                    "some code may not have run."
                )
                self.evt_summaryState.set_put(summaryState=self._summary_state)
            asyncio.create_task(self.handle_summary_state())
        finally:
            self._faulting = False

    def assert_enabled(self, action=""):
        """Assert that an action that requires ENABLED state can be run.

        Parameters
        ----------
        action : `str`, optional
            Action attempted. Not needed if this is called at the beginning
            of a ``do_...`` method, since the user will know what command
            was called.
        """
        if self.summary_state != State.ENABLED:
            what = f"{action} not allowed" if action else "Not allowed"
            if self.summary_state == State.FAULT:
                why = f"in Fault state; errorCode.errorReport={self.evt_errorCode.data.errorReport}"
            else:
                why = f"in state={self.summary_state!r}"
            raise base.ExpectedError(f"{what} {why}")

    @property
    def disabled_or_enabled(self):
        """Return True if the summary state is `State.DISABLED` or
        `State.ENABLED`.

        This is useful in `handle_summary_state` to determine if
        you should start or stop a telemetry loop,
        and connect to or disconnect from an external controller
        """
        return self.summary_state in (State.DISABLED, State.ENABLED)

    @property
    def summary_state(self):
        """Get the summary state as a `State` enum."""
        return self._summary_state

    async def handle_summary_state(self):
        """Called when the summary state has changed.

        Override to perform tasks such as starting and stopping telemetry
        (`example <https://ts-salobj.lsst.io/salobj_cscs.html
        #telemetry-loop-example>`_).

        Notes
        -----
        The versions in `BaseCsc` and `ConfigurableCsc` do nothing,
        so if you subclass one of those you do not need to call
        ``await super().handle_summary_state()``.
        """
        pass

    def report_summary_state(self):
        """Report a new value for summary_state, including current state.

        Subclasses may wish to override for code that depends on
        the current state (rather than the state transition command
        that got it into that state).
        """
        self.evt_summaryState.set_put(summaryState=self.summary_state)

    async def _do_change_state(self, data, cmd_name, allowed_curr_states, new_state):
        """Change to the desired state.

        Parameters
        ----------
        data : `DataType`
            Command data
        cmd_name : `str`
            Name of command, e.g. "disable" or "exitControl".
        allowed_curr_states : `list` of `State`
            Allowed current states
        new_state : `State`
            Desired new state.
        """
        curr_state = self.summary_state
        if curr_state not in allowed_curr_states:
            raise base.ExpectedError(f"{cmd_name} not allowed in state {curr_state!r}")
        try:
            await getattr(self, f"begin_{cmd_name}")(data)
        except base.ExpectedError as e:
            self.log.error(
                f"beg_{cmd_name} failed; remaining in state {curr_state!r}: {e}"
            )
            raise
        except Exception:
            self.log.exception(
                f"beg_{cmd_name} failed; remaining in state {curr_state!r}"
            )
            raise
        self._summary_state = new_state
        try:
            await getattr(self, f"end_{cmd_name}")(data)
        except base.ExpectedError as e:
            self.log.error(
                f"beg_{cmd_name} failed; reverting to state {curr_state!r}: {e}"
            )
            raise
        except Exception:
            self._summary_state = curr_state
            self.log.exception(
                f"end_{cmd_name} failed; reverting to state {curr_state!r}"
            )
            raise
        await self.handle_summary_state()
        self.report_summary_state()

    async def _heartbeat_loop(self):
        """Output heartbeat at regular intervals."""
        while True:
            try:
                await asyncio.sleep(self.heartbeat_interval)
                self.evt_heartbeat.put()
            except asyncio.CancelledError:
                break
            except Exception as e:
                # don't use the log because it also uses DDS messaging
                print(f"Heartbeat output failed: {e!r}", file=sys.stderr)



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["AsyncS3Bucket"]

import asyncio
import io
import os
import re

import astropy.time
import astropy.units as u
import boto3
import botocore
import moto


class AsyncS3Bucket:
    """Asynchronous interface to an Amazon Web Services s3 bucket.

    Parameters
    ----------
    name : `str`
        Name of bucket. If using Amazon Web Services see
        <https://docs.aws.amazon.com/AmazonS3/latest/dev/UsingBucket.html>
        for details. In particular note that bucket names must be globally
        unique across all AWS accounts.
    create : `bool`, optional
        If True and the bucket does not exist, create it.
        If False then assume the bucket exists.
        You will typically want true if using a mock server (``domock`` true).
    profile : `str`, optional
        Profile name; use the default profile if None.
    domock : `bool`, optional
        If True then start a mock S3 server.
        This is recommended for running in simulation mode.

    Attributes
    ----------
    service_resource : `boto3.resources.factory.s3.ServiceResource`
        The resource used to access the S3 service.
        Primarly provided for unit tests.
    name : `str`
        The bucket name.
    profile : `str`
        The profile name, or None if not specified.
    bucket : `boto3.resources.s3.Bucket`
        The S3 bucket.

    Notes
    -----
    Reads the following `Environment Variables
    <https://ts-salobj.lsst.io/configuration.html#environment_variables>`_;
    follow the link for details:

    * **S3_ENDPOINT_URL**:  The endpoint URL, e.g. ``http://foo.bar:9000``.

    The format for bucket names, file keys, and ``largeFileEvent`` event URLs
    is described in `CAP 452 <https://jira.lsstcorp.org/browse/CAP-452>`_
    """

    def __init__(self, name, *, create=False, profile=None, domock=False):
        self.mock = None
        self.profile = profile
        if domock:
            self._start_mock(name)

        endpoint_url = os.environ.get("S3_ENDPOINT_URL", None)
        if not endpoint_url:
            endpoint_url = None  # Handle ""

        session = boto3.Session(profile_name=profile)
        self.service_resource = session.resource("s3", endpoint_url=endpoint_url)
        self.name = name
        if create:
            # create_bucket is a no-op if the bucket already exists
            self.service_resource.create_bucket(Bucket=name)

        self.bucket = self.service_resource.Bucket(name)

    def _start_mock(self, name):
        """Start a mock S3 server with the specified bucket."""
        self.mock = moto.mock_s3()
        self.mock.start()

    def stop_mock(self):
        """Stop the mock s3 service, if running. A no-op if not running."""
        if self.mock is not None:
            self.mock.stop()
            self.mock = None

    @staticmethod
    def make_bucket_name(s3instance, s3category="LFA"):
        """Make an S3 bucket name.

        Parameters
        ----------
        salname : `str`
            SAL component name, e.g. ATPtg.
        salindexname : `str` or `None`
            For an indexed SAL component: a name associated with the SAL index,
            or just the index itself if there is no name.
            Specify `None` for a non-indexed SAL component.
            For example: "MT" for Main Telescope, "AT" for auxiliary telescope,
            or "ATBlue" for the AT Blue Fiber Spectrograph.
        s3instance : `str`
            S3 server instance. Typically "Summit", "Tucson" or "NCSA".
        s3category : `str`, optional
            Category of S3 server. The default is "LFA",
            for the Large File Annex.

        Returns
        -------
        bucket_name : `str`
            The S3 bucket name in the format described below:

        Raises
        ------
        ValueError
            If one or more arguments does not meet the rules below
            or the resulting bucket name is longer than 63 characters.

        Notes
        -----
        The rules for all arguments are as follows:

        * Each argument must start and end with a letter or digit.
        * Each argument may only contain letters, digits, and ".".

        The returned bucket name is cast to lowercase (because S3 bucket names
        may not contain uppercase letters) and has format::

            rubinobs-{s3category}-{s3instance}]
        """
        valid_arg_regex = re.compile(r"^[a-z0-9][a-z0-9.]*$", flags=re.IGNORECASE)
        kwargs = dict(s3instance=s3instance, s3category=s3category)
        for argname, arg in kwargs.items():
            if valid_arg_regex.match(arg) is None:
                raise ValueError(f"{argname}={arg} invalid")
            if valid_arg_regex.match(arg[-1:]) is None:
                raise ValueError(f"{argname}={arg} invalid")
        bucket_name = f"rubinobs-{s3category}-{s3instance}".lower()
        if len(bucket_name) > 63:
            raise ValueError(
                f"Bucket name {bucket_name!r} too long: len={len(bucket_name)} > 63 chars"
            )
        return bucket_name

    @staticmethod
    def make_key(salname, salindexname, generator, date, other=None, suffix=".dat"):
        """Make a key for an item of data.

        Parameters
        ----------
        salname : `str`
            SAL component name, e.g. ATPtg.
        salindexname : `str`, `int`, or `None`
            For an indexed SAL component: a name associated with the SAL index,
            or just the index itself if there is no name.
            Specify `None` for a non-indexed SAL component.
            For example: "MT" for Main Telescope, "AT" for auxiliary telescope,
            or "ATBlue" for the AT Blue Fiber Spectrograph.
        generator : `str`
            Dataset type.
        date : `astropy.time.Time`
            A date -- typically the date the data was taken.
        other : `str` or `None`, optional
            Additional text to identify the data and make the key unique.
            If `None` use ``date.tai.isot``: ``date`` as TAI,
            in ISO-8601 format, with a "T" between the date and time
            and a precision of milliseconds.
        suffix : `str`, optional
            Key suffix, e.g. ".fits".

        Returns
        -------
        key : `str`
            The key, as described below.

        Notes
        -----
        The returned key has format::

            {fullsalname}/{generator}/{yyyy}/{mm}/{dd}/
                {fullsalname}-{generator}-{other}{suffix}

        where:

        * ``fullsalname`` = ``{salname}:{salindexname}`` if ``salindexname``,
          else ``salname``.
        * ``yyyy``, ``mm``, ``dd`` are the "observing day":
          the year, month and day at TAI date - 12 hours,
          with 4, 2, 2 digits, respectively.
          The "observing day" does change during nighttime observing
          at the summit. Year, month and day are determined after rounding
          the date to milliseconds, so the reported observing day
          is consistent with the default value for ``other``.

        Note that the url field of the ``largeFileObjectAvailable`` event
        should have the format f"s3://{bucket}/{key}"
        """
        fullsalname = salname
        if salindexname:
            fullsalname += f":{salindexname}"
        taidate = astropy.time.Time(date, scale="tai", precision=3)
        shifted_isot = (taidate - 12 * u.hour).isot
        yyyy = shifted_isot[0:4]
        mm = shifted_isot[5:7]
        dd = shifted_isot[8:10]
        if other is None:
            other = taidate.isot
        return f"{fullsalname}/{generator}/{yyyy}/{mm}/{dd}/{fullsalname}_{generator}_{other}{suffix}"

    async def upload(self, fileobj, key, callback=None):
        """Upload a file-like object to the bucket.

        Parameters
        ----------
        fileobj : file-like object
            File-like object that can be read as *binary* data.
        key : `str`
            Name to use for the file in the bucket.
        callback : callable, optional
            Function to call with updates while writing. The function receives
            one argument: the number of bytes written. If the transfer is
            successful then it will always be called at least once,
            and the sum of the number of bytes for all calls will equal
            the size of the file.
            The callback function is called by ``S3.Bucket.upload_fileobj``.

        Notes
        -----
        To create a file-like object ``fileobj`` from an
        `astropy.io.fits.HDUList` named ``hdulist``::

            fileobj = io.BytesIO()
            hdulist.writeto(fileobj)
            fileobj.seek(0)
        """
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._sync_upload, fileobj, key, callback)

    async def download(self, key, callback=None):
        """Download a file-like object from the bucket.

        Parameters
        ----------
        key : `str`
            Name of the file in the bucket.
        callback : callable, optional
            Function to call with updates while writing. The function receives
            one argument: the number of bytes read so far. If the transfer is
            successful then it will always be called at least once,
            and the sum of the number of bytes for all calls will equal
            the size of the file.

        Returns
        -------
        fileobj : `io.BytesIO`
            The downloaded data as a file-like object.

        Notes
        -----
        To convert a file-like object ``fileobj`` to an
        `astropy.io.fits.HDUList` named ``hdulist``::

            hdulist = astropy.io.fits.open(fileobj)

        The callback function is called by ``S3.Bucket.download_fileobj``.
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._sync_download, key, callback)

    async def exists(self, key):
        """Check if a specified file exists in the bucket.

        Parameters
        ----------
        key : `str`
            Name of the potential file in the bucket.
        """
        loop = asyncio.get_running_loop()
        try:
            await loop.run_in_executor(None, self._sync_size, key)
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            # Something went wrong; report it.
            raise
        return True

    async def size(self, key):
        """Get the size in bytes of a given file in the bucket.

        Parameters
        ----------
        key : `str`
            Name of the file in the bucket.
        """
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._sync_size, key)

    def _sync_upload(self, fileobj, key, callback):
        self.bucket.upload_fileobj(Fileobj=fileobj, Key=key, Callback=callback)

    def _sync_download(self, key, callback):
        fileobj = io.BytesIO()
        self.bucket.download_fileobj(Key=key, Fileobj=fileobj, Callback=callback)
        # Rewind the fileobj so read returns the data.
        fileobj.seek(0)
        return fileobj

    def _sync_size(self, key):
        return self.bucket.meta.client.head_object(Bucket=self.name, Key=key)[
            "ContentLength"
        ]



"""Sphinx configuration file for an LSST stack package.

This configuration only affects single-package Sphinx documentation builds.
"""

from documenteer.conf.pipelinespkg import *  # noqa
import lsst.ts.salobj  # noqa

project = "ts_salobj"
html_theme_options["logotext"] = project  # noqa
html_title = project
html_short_title = project
doxylink = {}  # Avoid warning: Could not find tag file _doxygen/doxygen.tag



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["ControllerTelemetry"]

from . import write_topic


class ControllerTelemetry(write_topic.WriteTopic):
    """Write a specific telemetry topic.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Command name
    """

    def __init__(self, salinfo, name):
        super().__init__(salinfo=salinfo, name=name, sal_prefix="")

    def put(self, data=None):
        """Output this topic.

        Parameters
        ----------
        data : ``self.DataType``, optional
            New message data to replace ``self.data``, if any.

        Raises
        ------
        TypeError
            If ``data`` is not None and not an instance of `DataType`.
        """
        super().put(data=data)

    def set_put(self, **kwargs):
        """Set zero or more fields of ``self.data`` and put the result.

        Parameters
        ----------
        **kwargs : `dict` [`str`, ``any``]
            The keyword arguments are field name = new value for that field.
            See `set` for more information about values.

        Returns
        -------
        did_change : `bool`
            True if ``self.data`` was changed, or if this was the first call
            to `set`.

        Raises
        ------
        AttributeError
            If the topic does not have the specified field.
        ValueError
            If the field cannot be set to the specified value.
        """
        did_change = self.set(**kwargs)
        self.put()
        return did_change



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import sys
import unittest

import numpy as np

from lsst.ts import salobj

np.random.seed(47)

index_gen = salobj.index_generator()


class SimulationModeTestCase(unittest.IsolatedAsyncioTestCase):
    """Test simulation mode handling, including the --simulate
    command-line argument.
    """

    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        # Valid simulation modes that will exercise several things:
        # If 0 is present it is the default,
        # otherwise the first value is the default.
        # If 0 is present and there are two values then --simulate is a flag,
        # otherwise --simulate requires a value
        self.valid_simulation_modes_list = (
            (0, 1),
            (1, 0),
            (0, 3),
            (3, 0),
            (1, 2),
            (0, 1, 4),
            (4, 1, 0),
        )

    def make_csc_class(self, modes):
        """Make a subclass of TestCsc with specified valid simulation modes"""

        class TestCscWithSimulation(salobj.TestCsc):
            valid_simulation_modes = modes

        return TestCscWithSimulation

    async def test_invalid_simulation_modes(self):
        index = next(index_gen)
        for valid_simulation_modes in self.valid_simulation_modes_list:
            csc_class = self.make_csc_class(valid_simulation_modes)
            for bad_simulation_mode in range(-1, 5):
                if bad_simulation_mode in csc_class.valid_simulation_modes:
                    continue
                with self.assertRaises(ValueError):
                    csc_class(index=index, simulation_mode=bad_simulation_mode)

    async def test_valid_simulation_modes(self):
        for valid_simulation_modes in self.valid_simulation_modes_list:
            csc_class = self.make_csc_class(valid_simulation_modes)
            for simulation_mode in csc_class.valid_simulation_modes:
                index = next(index_gen)
                async with csc_class(
                    index=index, simulation_mode=simulation_mode
                ) as csc:
                    self.assertEqual(csc.simulation_mode, simulation_mode)

    async def test_simulate_cmdline_arg(self):
        orig_argv = sys.argv[:]
        try:
            for valid_simulation_modes in self.valid_simulation_modes_list:
                print(f"valid_simulation_modes={valid_simulation_modes}")
                if 0 in valid_simulation_modes:
                    default_simulation_mode = 0
                else:
                    default_simulation_mode = valid_simulation_modes[0]

                index = next(index_gen)
                csc_class = self.make_csc_class(valid_simulation_modes)
                if len(valid_simulation_modes) != 2 or 0 not in valid_simulation_modes:
                    # This should fail with no value
                    sys.argv = [
                        "test_csc.py",  # irrelevant
                        str(index),
                        "--simulate",
                    ]
                    with self.assertRaises(SystemExit):
                        csc_class.make_from_cmd_line(index=True)

                    # Test invalid simulation modes
                    for bad_simulation_mode in range(-1, 5):
                        if bad_simulation_mode in valid_simulation_modes:
                            continue
                        sys.argv = [
                            "test_csc.py",  # irrelevant
                            str(index),
                            "--simulate",
                            str(bad_simulation_mode),
                        ]
                        with self.assertRaises(SystemExit):
                            csc_class.make_from_cmd_line(index=True)

                    # Test valid simulation modes
                    for good_simulation_mode in valid_simulation_modes:
                        sys.argv = [
                            "test_csc.py",  # irrelevant
                            str(index),
                            "--simulate",
                            str(good_simulation_mode),
                        ]
                        csc = csc_class.make_from_cmd_line(index=True)
                        try:
                            # The simulation mode isn't assigned
                            # until the CSC starts.
                            await csc.start_task
                            self.assertEqual(csc.simulation_mode, good_simulation_mode)
                        finally:
                            await csc.do_exitControl(data=None)
                            await asyncio.wait_for(csc.done_task, timeout=5)

                else:
                    # This should fail with any value, valid or not
                    for simulation_mode in (0, 1):
                        sys.argv = [
                            "test_csc.py",  # irrelevant
                            str(index),
                            "--simulate",
                            str(simulation_mode),
                        ]
                        with self.assertRaises(SystemExit):
                            csc_class.make_from_cmd_line(index=True)

                    if valid_simulation_modes[0] == 0:
                        nondefault_simulation_mode = valid_simulation_modes[1]
                    else:
                        nondefault_simulation_mode = valid_simulation_modes[0]

                    index = next(index_gen)
                    sys.argv = [
                        "test_csc.py",  # irrelevant
                        str(index),
                        "--simulate",
                    ]
                    csc = csc_class.make_from_cmd_line(index=True)
                    try:
                        # The simulation mode isn't assigned
                        # until the CSC starts.
                        await csc.start_task
                        self.assertEqual(
                            csc.simulation_mode, nondefault_simulation_mode
                        )
                    finally:
                        await csc.do_exitControl(data=None)
                        await asyncio.wait_for(csc.done_task, timeout=5)

                # In all cases no --simulate arg should give the default.
                index = next(index_gen)
                sys.argv = [
                    "test_csc.py",  # irrelevant
                    str(index),
                ]
                csc = csc_class.make_from_cmd_line(index=True)
                try:
                    # The simulation mode isn't assigned until the CSC starts
                    await csc.start_task
                    self.assertEqual(csc.simulation_mode, default_simulation_mode)
                finally:
                    await csc.do_exitControl(data=None)
                    await asyncio.wait_for(csc.done_task, timeout=5)

        finally:
            sys.argv[:] = orig_argv

    async def test_none_valid_simulation_modes_simulation_mode(self):
        """Test that a CSC that uses the deprecated valid_simulation_modes=None
        checks simulation mode in start, not the constructor.

        The only valid simulation_mode is 0 because that's what BaseCsc
        supports and I didn't override that support.
        """
        TestCscWithDeprecatedSimulation = self.make_csc_class(None)

        self.assertEqual(TestCscWithDeprecatedSimulation.valid_simulation_modes, None)

        for bad_simulation_mode in (1, 2):
            index = next(index_gen)
            warning_regex = "valid_simulation_modes=None is deprecated"
            with self.assertWarnsRegex(DeprecationWarning, warning_regex):
                csc = TestCscWithDeprecatedSimulation(
                    index=index, simulation_mode=bad_simulation_mode
                )
            with self.assertRaises(salobj.base.ExpectedError):
                await csc.start_task

        # Test the one valid simulation mode
        index = next(index_gen)
        with self.assertWarnsRegex(DeprecationWarning, warning_regex):
            csc = TestCscWithDeprecatedSimulation(index=index, simulation_mode=0)
        try:
            await csc.start_task
            self.assertEqual(csc.simulation_mode, 0)
        finally:
            await csc.do_exitControl(data=None)
            await asyncio.wait_for(csc.done_task, timeout=5)

    async def test_none_valid_simulation_modes_cmdline(self):
        """Test that when valid_simulation_modes=None that the command
        parser does not add the --simulate argument.
        """
        TestCscWithDeprecatedSimulation = self.make_csc_class(None)

        self.assertEqual(TestCscWithDeprecatedSimulation.valid_simulation_modes, None)

        orig_argv = sys.argv[:]
        try:
            index = next(index_gen)
            # Try 0, the only valid value. This will still fail
            # because there is no --simulate command-line argument.
            simulation_mode = 0
            sys.argv = ["test_csc.py", str(index), "--simulate", str(simulation_mode)]
            with self.assertRaises(SystemExit):
                TestCscWithDeprecatedSimulation.make_from_cmd_line(index=True)
        finally:
            sys.argv[:] = orig_argv


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import pathlib
import unittest

import numpy as np
import yaml

from lsst.ts import salobj

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

np.random.seed(47)

TEST_DATA_DIR = TEST_CONFIG_DIR = pathlib.Path(__file__).resolve().parent / "data"
TEST_CONFIG_DIR = TEST_DATA_DIR / "config"


class ConfigurationTestCase(salobj.BaseCscTestCase, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        # defaults hard-coded in <ts_salobj_root>/schema/Test.yaml
        self.default_dict = dict(
            string0="default value for string0",
            bool0=True,
            int0=5,
            float0=3.14,
            intarr0=[-1, 1],
        )
        self.config_fields = self.default_dict.keys()

    def basic_make_csc(self, initial_state, config_dir, simulation_mode):
        return salobj.TestCsc(
            self.next_index(),
            initial_state=initial_state,
            config_dir=config_dir,
            simulation_mode=simulation_mode,
        )

    async def check_settings_events(self, config_file):
        """Check the settingsApplied and appliedSettingsMatchStart events.

        Parameters
        ----------
        config_file : `str`
            The name of the config file, or "" if none specified.
        """
        # settingsVersion.settingsApplied should start with
        # the config file name followed by a colon
        data = await self.assert_next_sample(
            topic=self.remote.evt_settingsApplied,
        )
        desired_prefix = config_file + ":"
        self.assertEqual(data.settingsVersion[: len(desired_prefix)], desired_prefix)

        # appliedSettingsMatchStartIsTrue.appliedSettingsMatchStartIsTrue
        # should be True after being configured.
        await self.assert_next_sample(
            topic=self.remote.evt_appliedSettingsMatchStart,
            appliedSettingsMatchStartIsTrue=True,
        )

    async def test_no_config_specified(self):
        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)

            expected_settings_url = pathlib.Path(TEST_CONFIG_DIR).resolve().as_uri()
            expected_settings_labels = ",".join(
                (
                    "all_fields",
                    "empty",
                    "some_fields",
                    "long_label1_in_an_attempt_to_make_recommendedSettingsLabels_go_over_256_chars",
                    "long_label2_in_an_attempt_to_make_recommendedSettingsLabels_go_over_256_chars",
                    "long_label3_in_an_attempt_to_make_recommendedSettingsLabels_go_over_256_chars",
                    "long_label4_in_an_attempt_to_make_recommendedSettingsLabels_go_over_256_chars",
                    "long_label5_in_an_attempt_to_make_recommendedSettingsLabels_go_over_256_chars",
                )
            )
            data = await self.assert_next_sample(
                topic=self.remote.evt_settingVersions,
                settingsUrl=expected_settings_url,
                recommendedSettingsLabels=expected_settings_labels,
            )
            self.assertTrue(len(data.recommendedSettingsVersion) > 0)
            await self.remote.cmd_start.start(timeout=STD_TIMEOUT)
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            for key, expected_value in self.default_dict.items():
                self.assertEqual(getattr(config, key), expected_value)

            # Test the softwareVersions event.
            # Assume the string length is the same for each field.
            metadata = self.csc.salinfo.metadata
            data = await self.assert_next_sample(
                topic=self.remote.evt_softwareVersions,
                xmlVersion=metadata.xml_version,
                salVersion=metadata.sal_version,
                openSpliceVersion=salobj.get_dds_version(),
                cscVersion=salobj.__version__,
            )
            await self.check_settings_events("")

    async def test_default_config_dir(self):
        async with self.make_csc(initial_state=salobj.State.STANDBY, config_dir=None):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            data = await self.remote.evt_settingVersions.next(
                flush=False, timeout=STD_TIMEOUT
            )
            self.assertTrue(len(data.recommendedSettingsVersion) > 0)
            self.assertEqual(data.settingsUrl[0:8], "file:///")
            config_path = pathlib.Path(data.settingsUrl[7:])
            self.assertTrue(config_path.samefile(self.csc.config_dir))

    async def test_empty_label(self):
        config_name = "empty"

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            await self.remote.cmd_start.set_start(
                settingsToApply=config_name, timeout=STD_TIMEOUT
            )
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            for key, expected_value in self.default_dict.items():
                self.assertEqual(getattr(config, key), expected_value)

            await self.check_settings_events("empty.yaml")

    async def test_some_fields_label(self):
        """Test a config with some fields set to valid values."""
        config_label = "some_fields"
        config_file = "some_fields.yaml"

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            await self.remote.cmd_start.set_start(
                settingsToApply=config_label, timeout=STD_TIMEOUT
            )
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            config_path = os.path.join(self.csc.config_dir, config_file)
            with open(config_path, "r") as f:
                config_yaml = f.read()
            config_from_file = yaml.safe_load(config_yaml)
            for key, default_value in self.default_dict.items():
                if key in config_from_file:
                    self.assertEqual(getattr(config, key), config_from_file[key])
                    self.assertNotEqual(getattr(config, key), default_value)
                else:
                    self.assertEqual(getattr(config, key), default_value)

            await self.check_settings_events(config_file)

    async def test_some_fields_file_no_hash(self):
        """Test a config specified by filename."""
        config_file = "some_fields.yaml"

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            await self.remote.cmd_start.set_start(
                settingsToApply=config_file, timeout=STD_TIMEOUT
            )
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            config_path = os.path.join(self.csc.config_dir, config_file)
            with open(config_path, "r") as f:
                config_yaml = f.read()
            config_from_file = yaml.safe_load(config_yaml)
            for key, default_value in self.default_dict.items():
                if key in config_from_file:
                    self.assertEqual(getattr(config, key), config_from_file[key])
                    self.assertNotEqual(getattr(config, key), default_value)
                else:
                    self.assertEqual(getattr(config, key), default_value)

            await self.check_settings_events(config_file)

    async def test_some_fields_file_with_hash(self):
        """Test a config specified by filename:hash."""
        config_file = "some_fields.yaml"

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            await self.remote.cmd_start.set_start(
                settingsToApply=f"{config_file}:HEAD", timeout=STD_TIMEOUT
            )
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            config_path = os.path.join(self.csc.config_dir, config_file)
            with open(config_path, "r") as f:
                config_yaml = f.read()
            config_from_file = yaml.safe_load(config_yaml)
            for key, default_value in self.default_dict.items():
                if key in config_from_file:
                    self.assertEqual(getattr(config, key), config_from_file[key])
                    self.assertNotEqual(getattr(config, key), default_value)
                else:
                    self.assertEqual(getattr(config, key), default_value)

            await self.check_settings_events(config_file)

    async def test_all_fields_label(self):
        """Test a config with all fields set to valid values."""
        config_name = "all_fields"
        config_file = "all_fields.yaml"

        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            await self.remote.cmd_start.set_start(
                settingsToApply=config_name, timeout=STD_TIMEOUT
            )
            await self.assert_next_summary_state(salobj.State.DISABLED)
            config = self.csc.config
            config_path = os.path.join(self.csc.config_dir, config_file)
            with open(config_path, "r") as f:
                config_yaml = f.read()
            config_from_file = yaml.safe_load(config_yaml)
            for key in self.config_fields:
                self.assertEqual(getattr(config, key), config_from_file[key])

            await self.check_settings_events(config_file)

    async def test_invalid_configs(self):
        async with self.make_csc(
            initial_state=salobj.State.STANDBY, config_dir=TEST_CONFIG_DIR
        ):
            await self.assert_next_summary_state(salobj.State.STANDBY)
            for name in ("all_bad_types", "bad_format", "one_bad_type", "extra_field"):
                config_file = f"invalid_{name}.yaml"
                with self.subTest(config_file=config_file):
                    with salobj.assertRaisesAckError(ack=salobj.SalRetCode.CMD_FAILED):
                        await self.remote.cmd_start.set_start(
                            settingsToApply=config_file, timeout=STD_TIMEOUT
                        )
                    data = self.remote.evt_summaryState.get()
                    self.assertEqual(self.csc.summary_state, salobj.State.STANDBY)
                    self.assertEqual(data.summaryState, salobj.State.STANDBY)

            # Make sure the CSC can still be started.
            await self.remote.cmd_start.set_start(
                settingsToApply="all_fields.yaml", timeout=10
            )
            self.assertEqual(self.csc.summary_state, salobj.State.DISABLED)
            await self.assert_next_summary_state(salobj.State.DISABLED)


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import enum
import os
import pathlib
import sys
import unittest

import numpy as np

from lsst.ts import salobj

np.random.seed(47)

index_gen = salobj.index_generator()
TEST_DATA_DIR = TEST_CONFIG_DIR = pathlib.Path(__file__).resolve().parent / "data"
TEST_CONFIG_DIR = TEST_DATA_DIR / "config"


class NoIndexCsc(salobj.TestCsc):
    """A CSC whose constructor has no index argument."""

    def __init__(
        self,
        arg1,
        arg2,
        initial_state=salobj.State.STANDBY,
        settings_to_apply="",
        config_dir=None,
    ):
        super().__init__(
            index=next(index_gen),
            initial_state=initial_state,
            settings_to_apply=settings_to_apply,
            config_dir=config_dir,
        )
        self.arg1 = arg1
        self.arg2 = arg2


class TestCscSettingsRequired(salobj.TestCsc):
    """A variant of TestCsc that has require_settings=True."""

    require_settings = True


class TestCscSettingsRequiredNoCmdLineState(salobj.TestCsc):
    """An invalid variant of TestCsc with require_settings=True
    and enable_cmdline_state=True.

    This should raise RuntimeError in make_from_cmd_line.
    """

    require_settings = True
    enable_cmdline_state = False


class CscMakeFromCmdLineTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()
        self.original_argv = sys.argv[:]

    def tearDown(self):
        sys.argv = self.original_argv

    async def test_no_index(self):
        for index in (0, None):
            with self.subTest(index=index):
                sys.argv = [sys.argv[0]]
                arg1 = "astring"
                arg2 = 2.75
                async with NoIndexCsc.make_from_cmd_line(
                    index=index, arg1=arg1, arg2=arg2
                ) as csc:
                    self.assertEqual(csc.arg1, arg1)
                    self.assertEqual(csc.arg2, arg2)
                    await csc.close()

    async def test_specified_index(self):
        sys.argv = [sys.argv[0]]
        index = next(index_gen)
        async with salobj.TestCsc.make_from_cmd_line(index=index) as csc:
            self.assertEqual(csc.salinfo.index, index)
            await csc.close()

    async def test_enum_index(self):
        class SalIndex(enum.IntEnum):
            ONE = 1
            TWO = 2

        for item in SalIndex:
            index = item.value
            sys.argv = [sys.argv[0], str(index)]
            async with salobj.TestCsc.make_from_cmd_line(index=SalIndex) as csc:
                self.assertEqual(csc.salinfo.index, index)
                await csc.close()

    async def test_index_from_argument_and_default_config_dir(self):
        index = next(index_gen)
        sys.argv = [sys.argv[0], str(index)]
        async with salobj.TestCsc.make_from_cmd_line(index=True) as csc:
            self.assertEqual(csc.salinfo.index, index)

            desired_config_pkg_name = "ts_config_ocs"
            desired_config_env_name = desired_config_pkg_name.upper() + "_DIR"
            desird_config_pkg_dir = os.environ[desired_config_env_name]
            desired_config_dir = pathlib.Path(desird_config_pkg_dir) / "Test/v1"
            self.assertEqual(csc.get_config_pkg(), desired_config_pkg_name)
            self.assertEqual(csc.config_dir, desired_config_dir)
            await csc.close()

    async def test_config_from_argument(self):
        index = next(index_gen)
        sys.argv = [sys.argv[0], str(index), "--config", str(TEST_CONFIG_DIR)]
        async with salobj.TestCsc.make_from_cmd_line(index=True) as csc:
            self.assertEqual(csc.salinfo.index, index)
            self.assertEqual(csc.config_dir, TEST_CONFIG_DIR)
            await csc.close()

    async def test_state_good(self):
        """Test valid --state and --settings arguments."""
        for state, settings in (
            (salobj.State.OFFLINE, ""),
            (salobj.State.STANDBY, ""),
            (salobj.State.DISABLED, "all_fields.yaml"),
            (salobj.State.ENABLED, ""),
        ):
            state_name = state.name.lower()
            index = next(index_gen)
            sys.argv = [
                sys.argv[0],
                str(index),
                "--config",
                str(TEST_CONFIG_DIR),
                "--state",
                state_name,
                "--settings",
                settings,
            ]
            async with salobj.TestCsc.make_from_cmd_line(index=True) as csc:
                self.assertEqual(csc.salinfo.index, index)
                self.assertEqual(csc.summary_state, state)
                if state not in (salobj.State.DISABLED, salobj.State.ENABLED):
                    # Configuration not applied.
                    self.assertIsNone(csc.config)
                elif settings == "all_fields.yaml":
                    self.assertEqual(csc.config.string0, "an arbitrary string")
                elif settings == "":
                    self.assertEqual(csc.config.string0, "default value for string0")
                else:
                    self.fail("Unhandled case.")
                await csc.close()

    async def test_state_bad(self):
        """Test invalid --state or --settings arguments."""
        # The only allowed --state values are "standby", "disabled",
        # and "enabled".
        for bad_state_name in ("fault", "not_a_state"):
            index = next(index_gen)
            sys.argv = [sys.argv[0], str(index), "--state", bad_state_name]
            with self.assertRaises(SystemExit):
                salobj.TestCsc.make_from_cmd_line(index=True)

        # Test that require_settings=True makes --settings required
        # if state is "disabled" or "enabled".
        for state_name in ("disabled", "enabled"):
            index = next(index_gen)
            sys.argv = [sys.argv[0], str(index), "--state", state_name]
            with self.assertRaises(SystemExit):
                TestCscSettingsRequired.make_from_cmd_line(index=True)

        # Test that require_settings=True requires enable_cmdline_state=True.
        # This is caught in the constructor, not the argument parser,
        # in order to make sure the developer sees it.
        index = next(index_gen)
        with self.assertRaises(RuntimeError):
            TestCscSettingsRequiredNoCmdLineState(index=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import argparse

from lsst.ts import salobj


class MinimalSalobjController(salobj.Controller):
    """Minimal Test controller using ts_salobj, for unit tests.

    Responds to one command: setLogLevel.

    Parameters
    ----------
    index : `int`
        SAL index.
    initial_log_level : `int`
        Initial log level.
    """

    def __init__(self, index, initial_log_level):
        print(
            f"SalobjController: starting with index={index}, "
            f"initial_log_level={initial_log_level}"
        )
        super().__init__(name="Test", index=index, do_callbacks=False)
        self.cmd_setLogLevel.callback = self.do_setLogLevel
        self.evt_logLevel.set(level=initial_log_level)
        self.tel_scalars.set(int0=initial_log_level)

    @classmethod
    def make_from_cmd_line(cls):
        """Make an instance from the command line."""
        parser = argparse.ArgumentParser("Run a minimal Salobj Test controller")
        parser.add_argument("index", type=int, help="Script SAL Component index")
        parser.add_argument("initial_log_level", type=int, help="Initial log level")
        args = parser.parse_args()
        return MinimalSalobjController(
            index=args.index, initial_log_level=args.initial_log_level
        )

    @classmethod
    async def amain(cls):
        """Make and run a controller."""
        controller = cls.make_from_cmd_line()
        await controller.done_task

    async def start(self):
        """Finish construction."""
        await self.salinfo.start()
        print(
            f"SalobjController: writing initial logLevel.level={self.evt_logLevel.data.level} event"
        )
        self.evt_logLevel.put()

    def do_setLogLevel(self, data):
        print(
            f"SalobjController: read setLogLevel(cmdid={data.private_seqNum}; "
            f"level={data.level})"
        )

        print(
            f"SalobjController: writing logLevel={data.level} event "
            "and the same value in scalars.int0 telemetry"
        )
        self.evt_logLevel.set_put(level=data.level)
        self.tel_scalars.set_put(int0=data.level)
        if data.level == 0:
            print("SalobjController: quitting")
            asyncio.create_task(self.close())


if __name__ == "__main__":
    asyncio.run(MinimalSalobjController.amain())
    print("SalobjController: done")



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["get_dds_version"]

import pathlib
import re

import dds


def get_dds_version(dds_file=None):
    """Get the version of OpenSplice dds library.

    If it cannot be determined, return "?".

    Parameters
    ----------
    dds_file : `str` or `None`, optional
        Value of ``dds.__file__``.
        Use `None` except when testing this function.
    """
    # In case ADLink adds the expected attribute...
    version = getattr(dds, "__version__", None)
    if version is not None:
        return version

    # Else try to get the version from the __file__ attribute
    if dds_file is None:
        dds_file = dds.__file__
    dds_path = pathlib.Path(dds_file)
    if len(dds_path.parts) < 2:
        return "?"
    parent_dir = dds_path.parts[-2]
    match = re.search(r"(\d+\.\d+\.\d+)", parent_dir)
    if match is None:
        return "?"
    return match.groups()[0]



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License

__all__ = ["BaseCscTestCase"]

import abc
import asyncio
import contextlib
import enum
import logging
import shutil
import subprocess

from . import base
from . import sal_enums
from . import testutils
from .domain import Domain
from .remote import Remote
from .csc_utils import get_expected_summary_states

# Standard timeout (sec)
# Long to avoid unnecessary timeouts on slow CI systems.
STD_TIMEOUT = 60


class BaseCscTestCase(metaclass=abc.ABCMeta):
    """Base class for CSC tests.

    Subclasses must:

    * Inherit both from this and `unittest.IsolatedAsyncioTestCase`.
    * Override `basic_make_csc` to return a CSC.

    Also we suggest:

    * Add a method ``test_standard_state_transitions`` which calls
      `check_standard_state_transitions`.
    * Add a method ``test_bin_script`` which calls `check_bin_script`,
      assuming you have a binary script to run your CSC.
    """

    _index_iter = base.index_generator()

    @abc.abstractmethod
    def basic_make_csc(self, initial_state, config_dir, simulation_mode, **kwargs):
        """Make and return a CSC.

        Parameters
        ----------
        initial_state : `lsst.ts.salobj.State` or `int`
            The initial state of the CSC.
        config_dir : `str`
            Directory of configuration files, or None for the standard
            configuration directory (obtained from
            `ConfigureCsc._get_default_config_dir`).
        simulation_mode : `int`
            Simulation mode.
        kwargs : `dict`
            Extra keyword arguments, if needed.
        """
        raise NotImplementedError()

    def next_index(self):
        return next(self._index_iter)

    @contextlib.asynccontextmanager
    async def make_csc(
        self,
        initial_state=sal_enums.State.STANDBY,
        config_dir=None,
        simulation_mode=0,
        log_level=None,
        timeout=STD_TIMEOUT,
        **kwargs,
    ):
        """Create a CSC and remote and wait for them to start,
        after setting a random $LSST_DDS_PARTITION_PREFIX.

        The csc is accessed as ``self.csc`` and the remote as ``self.remote``.

        Reads and checks all but the last ``summaryState`` event during
        startup.

        Parameters
        ----------
        name : `str`
            Name of SAL component.
        initial_state : `lsst.ts.salobj.State` or `int`, optional
            The initial state of the CSC. Defaults to STANDBY.
        config_dir : `str`, optional
            Directory of configuration files, or `None` (the default)
            for the standard configuration directory (obtained from
            `ConfigureCsc._get_default_config_dir`).
        simulation_mode : `int`, optional
            Simulation mode. Defaults to 0 because not all CSCs support
            simulation. However, tests of CSCs that support simulation
            will almost certainly want to set this nonzero.
        log_level : `int` or `None`, optional
            Logging level, such as `logging.INFO`.
            If `None` then do not set the log level, leaving the default
            behavior of `SalInfo`: increase the log level to INFO.
        timeout : `float`, optional
            Time limit for the CSC to start (seconds).
        **kwargs : `dict`, optional
            Extra keyword arguments for `basic_make_csc`.
            For a configurable CSC this may include ``settings_to_apply``,
            especially if ``initial_state`` is DISABLED or ENABLED.

        Notes
        -----
        Adds a logging.StreamHandler if one is not already present.
        """
        testutils.set_random_lsst_dds_partition_prefix()
        self.remote = None
        self.csc = None
        try:
            self.csc = self.basic_make_csc(
                initial_state=initial_state,
                config_dir=config_dir,
                simulation_mode=simulation_mode,
                **kwargs,
            )
            if len(self.csc.log.handlers) < 2:
                self.csc.log.addHandler(logging.StreamHandler())
            self.remote = Remote(
                domain=self.csc.domain,
                name=self.csc.salinfo.name,
                index=self.csc.salinfo.index,
            )
            if log_level is not None:
                self.csc.log.setLevel(log_level)

            await asyncio.wait_for(
                asyncio.gather(self.csc.start_task, self.remote.start_task),
                timeout=timeout,
            )

            if initial_state != self.csc.default_initial_state:
                # Check all expected summary states expect the final state.
                # That is omitted for backwards compatibility.
                expected_states = get_expected_summary_states(
                    initial_state=self.csc.default_initial_state,
                    final_state=initial_state,
                )[:-1]
                for state in expected_states:
                    await self.assert_next_summary_state(state)

            yield
        finally:
            if self.remote is not None:
                await self.remote.close()
            if self.csc is not None:
                await self.csc.close()

    async def assert_next_summary_state(
        self, state, flush=False, timeout=STD_TIMEOUT, remote=None
    ):
        """Wait for and check the next ``summaryState`` event.

        Parameters
        ----------
        state : `lsst.ts.salobj.State` or `int`
            Desired summary state.
        flush : `bool`, optional
            Flush the read queue before waiting?
        timeout : `float`, optional
            Time limit for getting the data sample (sec).
        remote : `Remote`, optional
            Remote to use; ``self.remote`` if None.
        """
        if remote is None:
            remote = self.remote
        await self.assert_next_sample(
            topic=remote.evt_summaryState,
            flush=flush,
            timeout=timeout,
            summaryState=state,
        )

    async def assert_next_sample(
        self, topic, flush=False, timeout=STD_TIMEOUT, **kwargs
    ):
        """Wait for the next data sample for the specified topic,
        check specified fields for equality, and return the data.

        Parameters
        ----------
        topic : `topics.ReadTopic`
            Topic to read, e.g. ``remote.evt_logMessage``.
        flush : `bool`, optional
            Flush the read queue before waiting?
        timeout : `double`, optional
            Time limit for getting the data sample (sec).
        kwargs : `dict`
            Dict of field_name: expected_value
            The specified fields will be checked for equality.

        Returns
        -------
        data : topic data type
            The data read.
        """
        data = await topic.next(flush=flush, timeout=timeout)
        for field_name, expected_value in kwargs.items():
            read_value = getattr(data, field_name, None)
            if read_value is None:
                self.fail(f"No such field {field_name} in topic {topic}")
            if isinstance(expected_value, enum.IntEnum):
                try:
                    read_value = type(expected_value)(read_value)
                except Exception:
                    pass
            self.assertEqual(
                read_value,
                expected_value,
                msg=f"Failed on field {field_name}: read {read_value!r} != expected {expected_value!r}",
            )
        return data

    async def check_bin_script(
        self,
        name,
        index,
        exe_name,
        default_initial_state=sal_enums.State.STANDBY,
        initial_state=None,
        settings_to_apply=None,
        cmdline_args=(),
        timeout=STD_TIMEOUT,
    ):
        """Test running the CSC command line script.

        Parameters
        ----------
        name : `str`
            Name of SAL component, e.g. "Rotator"
        index : `int` or `None`
            SAL index of component.
        exe_name : `str`
            Name of executable, e.g. "run_rotator.py"
        default_initial_state : `lsst.ts.salobj.State`, optional
            The default initial state of the CSC.
            Ignored unless `initial_state` is None.
        initial_state : `lsst.ts.salobj.State` or `int` or `None`, optional
            The desired initial state of the CSC; used to specify
            the ``--state`` command-argument.
        settings_to_apply : `str` or `None`, optional
            Value for the ``--settings`` command-line argument.
            Only relevant if ``initial_state`` is one of
            `salobj.State.DISABLED` or `salobj.State.ENABLED`.
        cmdline_args : `List` [`str`]
            Additional command-line arguments, such as "--simulate".
        timeout : `float`, optional
            Time limit for the CSC to start and output
            the summaryState event.
        """
        testutils.set_random_lsst_dds_partition_prefix()
        exe_path = shutil.which(exe_name)
        if exe_path is None:
            self.fail(
                f"Could not find bin script {exe_name}; did you setup or install this package?"
            )

        args = [exe_name]
        if index not in (None, 0):
            args += [str(index)]
        if initial_state is None:
            expected_states = [default_initial_state]
        else:
            args += ["--state", initial_state.name.lower()]
            expected_states = get_expected_summary_states(
                initial_state=default_initial_state,
                final_state=initial_state,
            )
        if settings_to_apply is not None:
            args += ["--settings", settings_to_apply]
        args += cmdline_args

        async with Domain() as domain, Remote(
            domain=domain, name=name, index=index
        ) as self.remote:
            print("check_bin_script running:", " ".join(args))
            process = await asyncio.create_subprocess_exec(
                *args,
                stderr=subprocess.PIPE,
            )
            try:
                for state in expected_states:
                    await self.assert_next_summary_state(state, timeout=timeout)
                if settings_to_apply is not None and settings_to_apply.endswith(
                    ".yaml"
                ):
                    # The string settings_to_apply is a file name, and so
                    # should appear in evt_settingsApplied.settingsVersion
                    data = await self.remote.evt_settingsApplied.next(
                        flush=False, timeout=STD_TIMEOUT
                    )
                    self.assertTrue(data.settingsVersion.startswith(settings_to_apply))
            finally:
                if process.returncode is None:
                    process.terminate()
                    await asyncio.wait_for(process.wait(), timeout=STD_TIMEOUT)
                else:
                    print("Warning: suprocess has already quit.")
                    try:
                        data = await process.stderr.read()
                        print("Subprocess stderr: ", data.decode())
                    except Exception as e:
                        print(f"Could not read subprocess stderr: {e}")

    async def check_standard_state_transitions(
        self,
        enabled_commands,
        skip_commands=None,
        settingsToApply="",
        timeout=STD_TIMEOUT,
    ):
        """Test standard CSC state transitions.

        Parameters
        ----------
        enabled_commands : `List` [`str`]
            List of CSC-specific commands that are valid in the enabled state.
            Need not include the standard commands, which are "disable"
            and "setLogLevel" (which is valid in any state).
        skip_commands : `List` [`str`] or `None`, optional
            List of commands to skip.
        settingsToApply : `str`, optional
            Value for the ``settingsToApply`` argument for the ``start``
            command.
        timeout : `float`, optional
            Time limit for state transition commands (seconds).

        Notes
        -----
        ``timeout`` is only used for state transition commands that
        are expected to succceed. ``STD_TIMEOUT`` is used for things
        that should happen quickly:

        * Commands that should fail, due to the CSC being in the wrong state.
        * The ``summaryState`` event after each state transition:
        """
        enabled_commands = tuple(enabled_commands)
        skip_commands = tuple(skip_commands) if skip_commands else ()

        # Start in STANDBY state.
        self.assertEqual(self.csc.summary_state, sal_enums.State.STANDBY)
        await self.assert_next_summary_state(sal_enums.State.STANDBY)
        await self.check_bad_commands(
            good_commands=("start", "exitControl", "setAuthList", "setLogLevel")
            + skip_commands
        )

        # Send start; new state is DISABLED.
        await self.remote.cmd_start.set_start(
            settingsToApply=settingsToApply, timeout=timeout
        )
        self.assertEqual(self.csc.summary_state, sal_enums.State.DISABLED)
        await self.assert_next_summary_state(sal_enums.State.DISABLED)
        await self.check_bad_commands(
            good_commands=("enable", "standby", "setAuthList", "setLogLevel")
            + skip_commands
        )

        # Send enable; new state is ENABLED.
        await self.remote.cmd_enable.start(timeout=timeout)
        self.assertEqual(self.csc.summary_state, sal_enums.State.ENABLED)
        await self.assert_next_summary_state(sal_enums.State.ENABLED)
        all_enabled_commands = tuple(
            sorted(
                set(("disable", "setAuthList", "setLogLevel")) | set(enabled_commands)
            )
        )
        await self.check_bad_commands(
            good_commands=all_enabled_commands + skip_commands
        )

        # Send disable; new state is DISABLED.
        await self.remote.cmd_disable.start(timeout=timeout)
        self.assertEqual(self.csc.summary_state, sal_enums.State.DISABLED)
        await self.assert_next_summary_state(sal_enums.State.DISABLED)

        # Send standby; new state is STANDBY.
        await self.remote.cmd_standby.start(timeout=timeout)
        self.assertEqual(self.csc.summary_state, sal_enums.State.STANDBY)
        await self.assert_next_summary_state(sal_enums.State.STANDBY)

        # Send exitControl; new state is OFFLINE.
        await self.remote.cmd_exitControl.start(timeout=timeout)
        self.assertEqual(self.csc.summary_state, sal_enums.State.OFFLINE)
        await self.assert_next_summary_state(sal_enums.State.OFFLINE)

    async def check_bad_commands(self, bad_commands=None, good_commands=None):
        """Check that bad commands fail.

        Parameters
        ----------
        bad_commands : `List`[`str`] or `None`, optional
            Names of bad commands to try, or None for all commands.
        good_commands : `List`[`str`] or `None`, optional
            Names of good commands to skip, or None to skip none.

        Notes
        -----
        If a command appears in both lists, it is considered a good command,
        so it is skipped.
        """
        if bad_commands is None:
            bad_commands = self.remote.salinfo.command_names
        if good_commands is None:
            good_commands = ()
        commands = self.remote.salinfo.command_names
        for command in commands:
            if command in good_commands:
                continue
            with self.subTest(command=command):
                cmd_attr = getattr(self.remote, f"cmd_{command}")
                with testutils.assertRaisesAckError(
                    ack=sal_enums.SalRetCode.CMD_FAILED
                ):
                    await cmd_attr.start(timeout=STD_TIMEOUT)



"""Make a Controller and Remote for the Test device
"""
from lsst.ts import salobj

salobj.set_random_lsst_dds_partition_prefix()
salinfo = salobj.SalInfo("Test", 1)
command_names = salinfo.command_names
event_names = salinfo.event_names
telemetry_names = salinfo.telemetry_names
print(f"commands for {salinfo.name}:")
for item in command_names:
    print(f"  {item}")
print(f"\nevents for {salinfo.name}:")
for item in event_names:
    print(f"  {item}")
print(f"\ntelemetry topics for {salinfo.name}:")
for item in telemetry_names:
    print(f"  {item}")

print("\nmake remote")
remote = salobj.Remote("Test", 1)
print("make controller")
controller = salobj.Controller("Test", 1)

print("\ndata fields for arrays")
print([item for item in dir(remote.cmd_setArrays.data) if not item.startswith("_")])

print("\ndata fields for scalars")
print([item for item in dir(remote.cmd_setScalars.data) if not item.startswith("_")])



import os
import sys
import pathlib
import setuptools

install_requires = [
    "numpy",
    "astropy",
    "boto3",
    "ts-dds",
    "ts-ddsconfig",
    "ts-idl>=1.2",
]
tests_require = install_requires + [
    "pytest",
    "pytest-cov",
    "pytest-flake8",
    "pytest-asyncio",
    "moto",
]
dev_requires = tests_require + ["documenteer[pipelines]"]
scm_version_template = """# Generated by setuptools_scm
__all__ = ["__version__"]

__version__ = "{version}"
"""

tools_path = pathlib.Path(setuptools.__path__[0])
base_prefix = pathlib.Path(sys.base_prefix)
data_files_path = tools_path.relative_to(base_prefix).parents[1]


def local_scheme(version):
    return ""


setuptools.setup(
    name="ts_salobj",
    description="Object-oriented interface to LSST SAL communications middleware.",
    use_scm_version={
        "write_to": "python/lsst/ts/salobj/version.py",
        "write_to_template": scm_version_template,
        "local_scheme": local_scheme,
    },
    include_package_data=True,
    setup_requires=["setuptools_scm", "pytest-runner"],
    install_requires=install_requires,
    package_dir={"": "python"},
    packages=setuptools.find_namespace_packages(where="python"),
    package_data={"": ["*.rst", "*.yaml"]},
    data_files=[(os.path.join(data_files_path, "schema"), ["schema/Test.yaml"])],
    scripts=["bin/run_test_csc.py", "bin/command_test_csc.py"],
    tests_require=tests_require,
    extras_require={"dev": dev_requires},
    license="GPL",
    project_urls={
        "Bug Tracker": "https://jira.lsstcorp.org/secure/Dashboard.jspa",
        "Source Code": "https://github.com/lsst-ts/ts_salobj",
    },
)



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License

__all__ = ["BaseConfigTestCase"]

import abc
import importlib
import os
import pathlib
import unittest
import yaml

import jsonschema

from . import validator


class BaseConfigTestCase(metaclass=abc.ABCMeta):
    """Base class for testing configuration files.

    Subclasses must:

    * Inherit both from this and `unittest.TestCase`.

    Also we suggest:

    * Add a method ``test_<foo>`` which calls
      `check_config_files`.
    """

    def get_module_dir(self, module_name):
        """Get the directory of a python module, by importing the module.

        Parameters
        ----------
        module_name : `str`
            Module name, e.g. "lsst.ts.salobj".

        Returns
        -------
        module_dir : `pathlib.Path`
            Module directory, e.g. ``<package_root>/lsst/ts/salobj``

        Raises
        ------
        ModuleNotFoundError
            If the module is not found.
        """
        module = importlib.import_module(module_name)
        init_path = pathlib.Path(module.__file__)
        return init_path.parent

    def get_schema(self, csc_package_root, sal_name=None, schema_subpath=None):
        """Get the config schema for a package, as a dict.

        The schema is expected to be:

            csc_package_root / "schema" / f"{sal_name}.yaml"

        Parameters
        ----------
        csc_package_root : `str` or `pathlib.Path`
            Root directory of CSC package.
        sal_name : `str`
            SAL component name, e.g. "Watcher".
            Ignored if ``schema_subpath`` is specified.
        schema_subpath : `str` or `None`
            Schema path relative to csc_package_root.
            If `None` then use ``f"schema/{sal_name}.yaml"``

        Raises
        ------
        AssertionError
            If csc_package_root is not an existing directory.
            If the schema file is not an existing file.
        jsonschema.exceptions.SchemaError
            If the file cannot be interpreted as a `dict`.
        """
        csc_package_root = pathlib.Path(csc_package_root)
        self.assertTrue(
            csc_package_root.is_dir(),
            f"csc_package_root={csc_package_root} is not a file",
        )

        if schema_subpath is None:
            if sal_name is None:
                raise RuntimeError("Must specify sal_name or schema_subpath")
            schema_path = csc_package_root / "schema" / f"{sal_name}.yaml"
        else:
            schema_path = csc_package_root / schema_subpath
        self.assertTrue(
            schema_path.is_file(), f"schema_path={schema_path} is not a file"
        )
        with open(schema_path, "r") as schema_file:
            schema_yaml = schema_file.read()
        schema = yaml.safe_load(schema_yaml)
        self.assertIsInstance(schema, dict)
        return schema

    def get_config_dir(self, config_package_root, sal_name, schema):
        """Get the directory of config files, assuming the standard
        ts_config_x package layout.

        The config dir is assumed to be as follows, where ``version``
        comes from the title field of the schema:

            config_package_root / sal_name / version

        Parameters
        ----------
        config_package_root : `str` or `pathlib.Path`
            Root directory of configuration package.
            For unit tests in a config package, this will work:

                config_package_root = pathlib.Path(__file__).parents[1]
        sal_name : `str`
            SAL component name, e.g. "Watcher".
        schema : `dict`
            Configuration schema.
            Used to determine the version.

        Returns
        -------
        config_dir : `pathlib.Path`
            Directory containing configuration files.
        """
        config_package_root = pathlib.Path(config_package_root)
        self.assertTrue(
            config_package_root.is_dir(),
            f"config_package_root={config_package_root} is not a dir",
        )

        version = schema["title"].split()[-1]
        self.assertTrue(
            version.startswith("v"),
            f"version={version} from schema title {schema['title']} does not start with 'v'",
        )

        config_dir = pathlib.Path(config_package_root) / sal_name / version
        self.assertTrue(config_dir.is_dir(), f"config_dir={config_dir} is not a dir")
        return config_dir

    def check_config_files(self, config_dir, schema):
        """Check all configuration files for a given package.

        Parameters
        ----------
        config_dir : `str` or `pathlib.Path`
            Directory containing config files and ``_labels.yaml``.
        schema : `dict`
            Configuration schema.
        """
        config_dir = pathlib.Path(config_dir)
        self.assertTrue(config_dir.is_dir(), f"config_dir={config_dir} is not a dir")
        self.assertIsInstance(schema, dict)

        config_validator = validator.DefaultingValidator(schema)
        version = schema["title"].split()[-1]
        self.assertTrue(
            version.startswith("v"),
            f"version={version} from schema title {schema['title']} does not start with 'v'",
        )

        config_files = config_dir.glob("*.yaml")

        # Check that all config files are valid.
        # Avoid using self.subTest because that somehow prevents
        # the following from working with known bad config files:
        #     with self.assertRaises(AssertionError): # or Exception
        #         self.check_config_files(...)
        bad_config_files = []
        for config_file in config_files:
            if config_file.name == "_labels.yaml":
                continue
            with open(config_file, "r") as f:
                config_yaml = f.read()
            try:
                config_dict = yaml.safe_load(config_yaml)
                # Delete metadata, if present
                if config_dict:
                    config_dict.pop("metadata", None)
                config_validator.validate(config_dict)
            except jsonschema.exceptions.ValidationError as e:
                print(f"config file {config_file} failed validation: {e}")
                bad_config_files.append(str(config_file))
        if bad_config_files:
            raise AssertionError(
                f"{len(bad_config_files)} config file(s) failed validation: {bad_config_files}"
            )

        # Check that all entries in the label dict are valid
        # and point to existing files.
        labels_path = config_dir / "_labels.yaml"
        self.assertTrue(
            labels_path.is_file(), f"labels file={labels_path} is not a file"
        )
        with open(labels_path, "r") as f:
            labels_yaml = f.read()
        if labels_yaml:
            input_dict = yaml.safe_load(labels_yaml)
            if input_dict is None:
                input_dict = {}
            else:
                self.assertIsInstance(input_dict, dict)
        else:
            input_dict = {}

        invalid_labels = []
        invalid_files = []
        missing_files = []
        for label, config_name in input_dict.items():
            if not label.isidentifier() or label.startswith("_"):
                invalid_labels.append(label)
            if config_name.strip().startswith("/") or not config_name.endswith(".yaml"):
                invalid_files.append(config_name)
            if not (config_dir / config_name).is_file():
                missing_files.append(config_name)
        if invalid_labels:
            self.fail(f"Invalid labels {invalid_labels} in {labels_path}")
        if invalid_files:
            self.fail(f"Labels to invalid file {invalid_files} in {labels_path}")
        if missing_files:
            self.fail(f"Labels to missing files {missing_files} in {labels_path}")

    def check_standard_config_files(
        self,
        sal_name=None,
        module_name=None,
        package_name=None,
        schema_name=None,
        schema_subpath=None,
        config_package_root=None,
        config_dir=None,
    ):
        """A wrapper around `check_config_files` that handles the most common
        case.

        Assumptions:

        * The module can be imported, if ``module_name`` specified,
          else environment variable ``f"{package_name.upper()}_DIR"`` exists.
          If not, skip the test, because the package is not available.
        * The schema is a module constant named ``schema_name``,
          if provided, else the schema is a file found as follows:

            * If ``module_name`` is provided,
              the module root is n+1 levels above the package,
              where n is the number of "." in ``module_name``
            * The schema is in ``module root / "schema" / f"{sal_name}.yaml"``
              unless you override this with ``schema_subpath``.

        Parameters
        ----------
        sal_name : `str`
            SAL component name, e.g. "Watcher".
        module_name : `str` or `None`, optional
            Module name, e.g. "lsst.ts.salobj".
            If not None then get the CSC package root by importing
            the module and ignore ``package_name``.
        package_name : `str`, optional
            If ``module_name`` is None then specify this argument
            and get the CSC package root using environment variable
            ``f"{package_name.upper()}_DIR"``.
            This is useful for non-python packages or packages that
            need a special environment to be imported.
        schema_name : `str` or None, optional
            Name of schema constant in the module, typically "CONFIG_SCHEMA".
            If None then look for a schema file instead of a schema constant.
        schema_subpath : `str` or `None`
            Schema path relative to csc_package_root.
            If `None` then use ``f"schema/{sal_name}.yaml"``.
            Ignored if ``schema_name`` is specified.
        config_package_root : `str` or `pathlib.Path` or `None`
            Root directory of configuration package.
            Within the unit test this will work:

                config_package_root = pathlib.Path(__file__).parents[1]

            Ignored if ``schema_name`` or ``config_dir`` is specified.

        config_dir : `str` or `pathlib.Path` or `None`
            Directory containing config files and ``_labels.yaml``.
            If `None` then a reasonable value is computed;
            this is primarily intended to support unit testing in ts_salobj.
        """
        if module_name is not None and schema_name is not None:
            try:
                module = importlib.import_module(module_name)
            except ImportError:
                raise unittest.SkipTest(f"Cannot import {module_name}")
            schema = getattr(module, schema_name)
        else:
            if module_name is None:
                if package_name is None:
                    raise RuntimeError("Must specify module_name or package_name")
                else:
                    env_var = package_name.upper() + "_DIR"
                    csc_package_root = os.environ.get(env_var)
                    if csc_package_root is None:
                        raise unittest.SkipTest(
                            f"Cannot find package {package_name}; "
                            f"environment variable {env_var} not defined"
                        )
            else:
                num_dots = len(module_name.split("."))
                try:
                    csc_package_root = self.get_module_dir(module_name).parents[
                        num_dots
                    ]
                except ImportError:
                    raise unittest.SkipTest(f"Cannot import {module_name}")
            schema = self.get_schema(
                csc_package_root=csc_package_root,
                sal_name=sal_name,
                schema_subpath=schema_subpath,
            )
        if config_dir is None:
            if config_package_root is None:
                raise RuntimeError(
                    "config_package_root or config_dir must be specified."
                )
            config_dir = self.get_config_dir(
                config_package_root=config_package_root,
                sal_name=sal_name,
                schema=schema,
            )
        self.check_config_files(config_dir=config_dir, schema=schema)



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["MAX_SEQ_NUM", "WriteTopic"]

import struct

import numpy as np

from .base_topic import BaseTopic
from .. import base

# Maximum value for the ``private_seqNum`` field of each topic,
# a 4 byte signed integer.
# For command topics this field is the command ID, and it must be unique
# for each command in order to avoid collisions (since there is only one
# ``ackcmd`` topic that is shared by all commands).
# For other topics its use is unspecified but it may prove handy to
# increment it (with wraparound) for each data point.
MAX_SEQ_NUM = (1 << 31) - 1


class WriteTopic(BaseTopic):
    r"""Base class for topics that are output.

    This includes  controller events, controller telemetry, remote commands
    and ``cmdack`` writers.

    Parameters
    ----------
    salinfo : `.SalInfo`
        SAL component information
    name : `str`
        Topic name, without a "command\_" or "logevent\_" prefix.
    sal_prefix : `str`
        SAL topic prefix: one of "command\_", "logevent\_" or ""
    min_seq_num : `int` or `None`, optional
        Minimum value for the ``private_seqNum`` field. The default is 1.
        If `None` then ``private_seqNum`` is not set; this is needed
        for the cmdack writer, which sets the field itself.
    max_seq_num : `int`, optional
        Maximum value for ``private_seqNum``, inclusive.
        The default is the maximum allowed positive value
        (``private_seqNum`` is a 4-byte signed int).
        Ignored if ``min_seq_num`` is `None`.
    initial_seq_num : `int`, optional
        Initial sequence number; if `None` use min_seq_num.

    Attributes
    ----------
    isopen : `bool`
        Is this read topic open? `True` until `close` is called.
    """

    def __init__(
        self,
        *,
        salinfo,
        name,
        sal_prefix,
        min_seq_num=1,
        max_seq_num=MAX_SEQ_NUM,
        initial_seq_num=None,
    ):
        super().__init__(
            salinfo=salinfo,
            name=name,
            sal_prefix=sal_prefix,
        )
        self.isopen = True
        self.min_seq_num = min_seq_num  # record for unit tests
        self.max_seq_num = max_seq_num
        if min_seq_num is None:
            self._seq_num_generator = None
        else:
            self._seq_num_generator = base.index_generator(
                imin=min_seq_num, imax=max_seq_num, i0=initial_seq_num
            )
        # Command topics use a different a partition name than
        # all other topics, including ackcmd, and the partition name
        # is part of the publisher and subscriber.
        # This split allows us to create just one subscriber and one publisher
        # for each Controller or Remote:
        # `Controller` only needs a cmd_subscriber and data_publisher,
        # `Remote` only needs a cmd_publisher and data_subscriber.
        if sal_prefix == "command_":
            publisher = salinfo.cmd_publisher
        else:
            publisher = salinfo.data_publisher
        self._writer = publisher.create_datawriter(self._topic, self.qos_set.writer_qos)
        self._has_data = False
        self._data = self.DataType()
        self._has_priority = sal_prefix == "logevent_"

        salinfo.add_writer(self)

    @property
    def data(self):
        """Internally cached message.

        Raises
        ------
        TypeError
            If ``data`` is not an instance of `DataType`

        Notes
        -----
        Do not assume the data will be constant. You can make a copy
        using ``copy.copy(data)``.
        """
        return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, self.DataType):
            raise TypeError(f"data={data!r} must be an instance of {self.DataType}")
        self._data = data
        self._has_data = True

    @property
    def has_data(self):
        """Has `data` ever been set?"""
        return self._has_data

    def basic_close(self):
        """A synchronous and possibly less thorough version of `close`.

        Intended for exit handlers and constructor error handlers.
        """
        if not self.isopen:
            return
        self.isopen = False
        self._writer.close()

    async def close(self):
        """Shut down and release resources.

        Intended to be called by SalInfo.close(),
        since that tracks all topics.
        """
        self.basic_close()

    def put(self, data=None, priority=0):
        """Output this topic.

        Parameters
        ----------
        data : ``self.DataType`` or `None`
            New message data to replace ``self.data``, if any.
        priority : `int`, optional
            Priority; used to set the priority field of events.
            Ignored for commands and telemetry.

        Raises
        ------
        TypeError
            If ``data`` is not None and not an instance of `DataType`.
        """
        if data is not None:
            self.data = data

        if self._has_priority:
            self.data.priority = priority
        self.data.private_sndStamp = base.current_tai()
        self.data.private_revCode = self.rev_code
        self.data.private_origin = self.salinfo.domain.origin
        self.data.private_identity = self.salinfo.identity
        if self._seq_num_generator is not None:
            self.data.private_seqNum = next(self._seq_num_generator)
        # when index is 0 use the default of 0 and give senders a chance
        # to override it.
        if self.salinfo.index != 0:
            setattr(self.data, f"{self.salinfo.name}ID", self.salinfo.index)
        try:
            self._writer.write(self.data)
        except struct.error as e:
            raise ValueError(
                f"{self.name} write({self.data}) failed: one or more fields invalid"
            ) from e
        except TypeError as e:
            raise ValueError(
                f"{self.name} write({self.data}) failed: "
                f"perhaps one or more array fields has been set to a scalar"
            ) from e
        except IndexError as e:
            raise ValueError(
                f"{self.name} write({self.data}) failed: "
                f"probably one or more array fields is too short"
            ) from e

    def set(self, **kwargs):
        """Set one or more fields of message data cache ``self.data``.

        Parameters
        ----------
        **kwargs : `dict` [`str`, ``any``]
            Dict of field name: new value for that field:

            * Any key whose value is `None` is checked for existence,
              but the value of the field is not changed.
            * If the field being set is an array then the value must either
              be an array of the same length or a scalar (which replaces
              every element of the array).

        Returns
        -------
        did_change : `bool`
            True if ``self.data`` was changed, or if this was the first call
            to `set`.

        Raises
        ------
        AttributeError
            If the topic does not have the specified field.
        ValueError
            If the field cannot be set to the specified value.

        Notes
        -----
        If one or more fields cannot be set, the message data may be
        partially updated.
        """
        did_change = not self.has_data
        for field_name, value in kwargs.items():
            if value is None:
                if not hasattr(self.data, field_name):
                    raise AttributeError(f"{self.data} has no attribute {field_name}")
                continue
            old_value = getattr(self.data, field_name)
            try:
                if not did_change:
                    # array_equal works for sequences of all kinds,
                    # as well as strings and scalars
                    did_change |= not np.array_equal(old_value, value)
                setattr(self.data, field_name, value)
            except Exception as e:
                raise ValueError(
                    f"Could not set {self.data}.{field_name} to {value!r}"
                ) from e
        self._has_data = True
        return did_change



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import os
import unittest

import numpy as np

from lsst.ts import salobj

# Long enough to perform any reasonable operation
# including starting a CSC or loading a script (seconds)
STD_TIMEOUT = 60

HISTORY_TIMEOUT_NAME = "LSST_DDS_HISTORYSYNC"
INITIAL_HISTORY_TIMEOUT = os.environ.get(HISTORY_TIMEOUT_NAME, None)

np.random.seed(47)

index_gen = salobj.index_generator()


class RemoteTestCase(unittest.IsolatedAsyncioTestCase):
    async def test_constructor_include_exclude(self):
        """Test the include and exclude arguments for salobj.Remote."""

        index = next(index_gen)
        async with salobj.Domain() as domain:
            salinfo = salobj.SalInfo(domain=domain, name="Test", index=index)

            # all possible expected topic names
            all_command_names = set(salinfo.command_names)
            all_event_names = set(salinfo.event_names)
            all_telemetry_names = set(salinfo.telemetry_names)

            # the associated method names
            all_command_method_names = set(f"cmd_{name}" for name in all_command_names)
            all_event_method_names = set(f"evt_{name}" for name in all_event_names)
            all_telemetry_method_names = set(
                f"tel_{name}" for name in all_telemetry_names
            )

            # remote0 specifies neither include nor exclude;
            # it should have everything
            remote0 = salobj.Remote(domain=domain, name="Test", index=index)
            remote_command_names = set(
                [name for name in dir(remote0) if name.startswith("cmd_")]
            )
            self.assertEqual(remote_command_names, all_command_method_names)
            remote_event_names = set(
                [name for name in dir(remote0) if name.startswith("evt_")]
            )
            self.assertEqual(remote_event_names, all_event_method_names)
            remote_telemetry_names = set(
                [name for name in dir(remote0) if name.startswith("tel_")]
            )
            self.assertEqual(remote_telemetry_names, all_telemetry_method_names)

            # remote1 uses the include argument
            include = ["errorCode", "scalars"]
            remote1 = salobj.Remote(
                domain=domain, name="Test", index=index, include=include
            )
            remote1_command_names = set(
                [name for name in dir(remote1) if name.startswith("cmd_")]
            )
            self.assertEqual(remote1_command_names, all_command_method_names)
            remote1_event_names = set(
                [name for name in dir(remote1) if name.startswith("evt_")]
            )
            self.assertEqual(
                remote1_event_names,
                set(f"evt_{name}" for name in include if name in all_event_names),
            )
            remote1_telemetry_names = set(
                [name for name in dir(remote1) if name.startswith("tel_")]
            )
            self.assertEqual(
                remote1_telemetry_names,
                set(f"tel_{name}" for name in include if name in all_telemetry_names),
            )

            # remote2 uses the exclude argument
            exclude = ["appliedSettingsMatchStart", "arrays"]
            remote2 = salobj.Remote(
                domain=domain, name="Test", index=index, exclude=exclude
            )
            remote2_command_names = set(
                [name for name in dir(remote2) if name.startswith("cmd_")]
            )
            self.assertEqual(remote2_command_names, all_command_method_names)
            remote2_event_names = set(
                [name for name in dir(remote2) if name.startswith("evt_")]
            )
            self.assertEqual(
                remote2_event_names,
                set(f"evt_{name}" for name in all_event_names if name not in exclude),
            )
            remote2_telemetry_names = set(
                [name for name in dir(remote2) if name.startswith("tel_")]
            )
            self.assertEqual(
                remote2_telemetry_names,
                set(
                    f"tel_{name}" for name in all_telemetry_names if name not in exclude
                ),
            )

            # remote3 omits commands
            remote3 = salobj.Remote(
                domain=domain, name="Test", index=index, readonly=True
            )
            remote_command_names = set(
                [name for name in dir(remote3) if name.startswith("cmd_")]
            )
            self.assertEqual(remote_command_names, set())
            remote_event_names = set(
                [name for name in dir(remote3) if name.startswith("evt_")]
            )
            self.assertEqual(remote_event_names, all_event_method_names)
            remote_telemetry_names = set(
                [name for name in dir(remote3) if name.startswith("tel_")]
            )
            self.assertEqual(remote_telemetry_names, all_telemetry_method_names)

            # remote4 uses include=[]
            remote4 = salobj.Remote(domain=domain, name="Test", index=index, include=[])
            remote_command_names = set(
                [name for name in dir(remote4) if name.startswith("cmd_")]
            )
            self.assertEqual(remote_command_names, all_command_method_names)
            remote_event_names = set(
                [name for name in dir(remote4) if name.startswith("evt_")]
            )
            self.assertEqual(remote_event_names, set())
            remote_telemetry_names = set(
                [name for name in dir(remote4) if name.startswith("tel_")]
            )
            self.assertEqual(remote_telemetry_names, set())

            # remote5 uses exclude=[] (though there is no reason to doubt
            # that it will work the same as exclude=None)
            remote5 = salobj.Remote(domain=domain, name="Test", index=index, exclude=[])
            remote_command_names = set(
                [name for name in dir(remote5) if name.startswith("cmd_")]
            )
            self.assertEqual(remote_command_names, all_command_method_names)
            remote_event_names = set(
                [name for name in dir(remote5) if name.startswith("evt_")]
            )
            self.assertEqual(remote_event_names, all_event_method_names)
            remote_telemetry_names = set(
                [name for name in dir(remote5) if name.startswith("tel_")]
            )
            self.assertEqual(remote_telemetry_names, all_telemetry_method_names)

            # make sure one cannot specify both include and exclude
            with self.assertRaises(ValueError):
                salobj.Remote(
                    domain=domain,
                    name="Test",
                    index=index,
                    include=include,
                    exclude=exclude,
                )

    def assert_max_history(self, remote, evt_max_history=1):
        for evt in [
            getattr(remote, f"evt_{name}") for name in remote.salinfo.event_names
        ]:
            self.assertEqual(evt.max_history, evt_max_history)

        for tel in [
            getattr(remote, f"tel_{name}") for name in remote.salinfo.telemetry_names
        ]:
            self.assertEqual(tel.max_history, 0)
            self.assertFalse(evt.volatile)

    async def test_default_max_history(self):
        """Test default evt_max_history ctor argument."""
        index = next(index_gen)
        async with salobj.Domain() as domain:
            remote = salobj.Remote(domain=domain, name="Test", index=index)
            self.assert_max_history(remote)

    async def test_evt_max_history(self):
        """Test non-default evt_max_history Remote constructor argument."""
        evt_max_history = 0
        index = next(index_gen)
        async with salobj.Domain() as domain:
            remote = salobj.Remote(
                domain=domain, name="Test", index=index, evt_max_history=evt_max_history
            )
            self.assert_max_history(remote, evt_max_history=evt_max_history)

    async def test_start_false(self):
        """Test the start argument of Remote."""

        index = next(index_gen)
        async with salobj.Domain() as domain:
            remote = salobj.Remote(domain=domain, name="Test", index=index, start=False)
            self.assertFalse(hasattr(remote, "start_task"))

    @unittest.skipIf(
        INITIAL_HISTORY_TIMEOUT is not None and float(INITIAL_HISTORY_TIMEOUT) < 0,
        f"${HISTORY_TIMEOUT_NAME}={INITIAL_HISTORY_TIMEOUT} < 0",
    )
    async def test_negative_lsst_dds_historysync(self):
        """Test that setting LSST_DDS_HISTORYSYNC < 0
        prevents waiting for historical data.

        This setting applies to SalInfo, not Remote,
        but it requires some topics in order to be tested,
        and Remote provides topics.
        """
        index = next(index_gen)
        async with salobj.Domain() as domain:
            # Make a normal remote that waits for historical data.
            remote1 = salobj.Remote(domain=domain, name="Test", index=index)
            await asyncio.wait_for(remote1.start_task, timeout=STD_TIMEOUT)
            self.assertGreater(len(remote1.salinfo.wait_history_isok), 0)

            # Make a remote that does not wait for historical data
            # by defining the history timeout env variable < 0.
            os.environ[HISTORY_TIMEOUT_NAME] = "-1"
            try:
                remote2 = salobj.Remote(domain=domain, name="Test", index=index)
                await asyncio.wait_for(remote2.start_task, timeout=STD_TIMEOUT)
            finally:
                if INITIAL_HISTORY_TIMEOUT is None:
                    del os.environ[HISTORY_TIMEOUT_NAME]
                else:
                    os.environ[HISTORY_TIMEOUT_NAME] = INITIAL_HISTORY_TIMEOUT

            self.assertEqual(len(remote2.salinfo.wait_history_isok), 0)


if __name__ == "__main__":
    unittest.main()



# This file is part of ts_salobj.
#
# Developed for the Rubin Observatory Telescope and Site System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import getpass
import os
import random
import socket
import time
import unittest

import astropy.time
from astropy.coordinates import Angle
import astropy.units as u
import numpy as np

from lsst.ts import salobj


def alternate_tai_from_utc_unix(utc_unix):
    """Compute TAI in unix seconds given UTC in unix seconds.

    Parameters
    ----------
    utc_unix : `float`

    This is not the implementation in base.py for two reasons:

    * It is too slow; it slows tests/test_speed.py by a factor of 8.
    * It blocks while downloading a leap second table.
    """
    ap_time = astropy.time.Time(utc_unix, scale="utc", format="unix")
    return ap_time.tai.mjd * salobj.SECONDS_PER_DAY - salobj.MJD_MINUS_UNIX_SECONDS


class BasicsTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        salobj.set_random_lsst_dds_partition_prefix()

    async def test_assert_raises_ack_error(self):
        """Test the assertRaisesAckError function."""
        async with salobj.Domain() as domain:
            salinfo = salobj.SalInfo(domain, "Test", index=1)
            private_seqNum = 5
            ack = 23
            error = -6
            result = "a result"
            err = salobj.AckError(
                "a message",
                ackcmd=salinfo.make_ackcmd(
                    private_seqNum=private_seqNum, ack=ack, error=error, result=result
                ),
            )
            self.assertEqual(err.ackcmd.private_seqNum, private_seqNum)
            self.assertEqual(err.ackcmd.ack, ack)
            self.assertEqual(err.ackcmd.error, error)
            self.assertEqual(err.ackcmd.result, result)

            for ExceptionClass in (
                Exception,
                TypeError,
                KeyError,
                RuntimeError,
                AssertionError,
            ):
                with self.assertRaises(ExceptionClass):
                    with salobj.assertRaisesAckError():
                        raise ExceptionClass(
                            "assertRaisesAckError should ignore other exception types"
                        )

            with self.assertRaises(AssertionError):
                with salobj.assertRaisesAckError(ack=5):
                    raise salobj.AckError(
                        "mismatched ack",
                        ackcmd=salinfo.make_ackcmd(private_seqNum=1, ack=1),
                    )

            with self.assertRaises(AssertionError):
                with salobj.assertRaisesAckError(error=47):
                    raise salobj.AckError(
                        "mismatched error",
                        ackcmd=salinfo.make_ackcmd(private_seqNum=2, ack=25, error=2),
                    )

            with salobj.assertRaisesAckError():
                raise salobj.AckError(
                    "no ack or error specified",
                    ackcmd=salinfo.make_ackcmd(private_seqNum=3, ack=1, error=2),
                )

            result = "result for this exception"
            # test result_contains with the full result string
            with salobj.assertRaisesAckError(ack=1, error=2, result_contains=result):
                raise salobj.AckError(
                    "match ack, error and full result",
                    ackcmd=salinfo.make_ackcmd(
                        private_seqNum=4, ack=1, error=2, result=result
                    ),
                )
            # test result_contains with a substring of the result string
            with salobj.assertRaisesAckError(
                ack=1, error=2, result_contains=result[2:-2]
            ):
                raise salobj.AckError(
                    "match ack, error and a substring of result",
                    ackcmd=salinfo.make_ackcmd(
                        private_seqNum=4, ack=1, error=2, result=result
                    ),
                )

    async def test_ack_error_repr(self):
        """Test AckError.__str__ and AckError.__repr__"""
        async with salobj.Domain() as domain:
            salinfo = salobj.SalInfo(domain, "Test", index=1)
            msg = "a message"
            private_seqNum = 5
            ack = 23
            error = -6
            result = "a result"
            err = salobj.AckError(
                msg,
                ackcmd=salinfo.make_ackcmd(
                    private_seqNum=private_seqNum, ack=ack, error=error, result=result
                ),
            )
            str_err = str(err)
            for item in (msg, private_seqNum, ack, error, result):
                self.assertIn(str(item), str_err)
            self.assertNotIn("AckError", str_err)
            repr_err = repr(err)
            for item in ("AckError", msg, private_seqNum, ack, error, result):
                self.assertIn(str(item), repr_err)

    def test_astropy_time_from_tai_unix(self):
        # Check the function at a leap second transition,
        # since that is likely to cause problems
        unix_time0 = datetime.datetime.fromisoformat("2017-01-01").timestamp()
        for dt in (-1, -0.5, -0.1, 0, 0.1, 1):
            with self.subTest(dt=dt):
                utc_unix = unix_time0 + dt
                tai_unix = salobj.tai_from_utc(utc_unix)
                astropy_time1 = salobj.astropy_time_from_tai_unix(tai_unix)
                self.assertIsInstance(astropy_time1, astropy.time.Time)
                self.assertEqual(astropy_time1.scale, "tai")
                tai_unix_round_trip1 = salobj.tai_from_utc(astropy_time1)
                self.assertAlmostEqual(tai_unix, tai_unix_round_trip1, delta=1e-6)

    async def test_get_opensplice_version(self):
        ospl_version = salobj.get_opensplice_version()
        self.assertRegex(ospl_version, r"^\d+\.\d+\.\d+")

    async def test_get_user_host(self):
        expected_user_host = getpass.getuser() + "@" + socket.getfqdn()
        user_host = salobj.get_user_host()
        self.assertEqual(expected_user_host, user_host)

    async def test_long_ack_result(self):
        async with salobj.Domain() as domain:
            salinfo = salobj.SalInfo(domain, "Test", index=1)
            ack = salobj.SalRetCode.CMD_FAILED
            error = 15
            long_result = (
                "this string is longer than MAX_RESULT_LEN characters "
                "this string is longer than MAX_RESULT_LEN characters "
                "this string is longer than MAX_RESULT_LEN characters "
                "this string is longer than MAX_RESULT_LEN characters "
                "this string is longer than MAX_RESULT_LEN characters "
            )
            self.assertGreater(len(long_result), salobj.MAX_RESULT_LEN)
            with self.assertRaises(ValueError):
                salinfo.make_ackcmd(
                    private_seqNum=1,
                    ack=ack,
                    error=error,
                    result=long_result,
                    truncate_result=False,
                )
            ackcmd = salinfo.make_ackcmd(
                private_seqNum=2,
                ack=ack,
                error=error,
                result=long_result,
                truncate_result=True,
            )
            self.assertEqual(ackcmd.result, long_result[0 : salobj.MAX_RESULT_LEN])
            self.assertEqual(ackcmd.ack, ack)
            self.assertEqual(ackcmd.error, error)

    def test_set_random_lsst_dds_domain(self):
        """Test that set_random_lsst_dds_domain is a deprecated
        alias for set_random_lsst_dds_partition_prefix.
        """
        old_prefix = os.environ["LSST_DDS_PARTITION_PREFIX"]
        with self.assertWarnsRegex(
            DeprecationWarning, "Use set_random_lsst_dds_partition_prefix"
        ):
            salobj.set_random_lsst_dds_domain()
        new_prefix = os.environ["LSST_DDS_PARTITION_PREFIX"]
        self.assertNotEqual(old_prefix, new_prefix)

    def test_set_random_lsst_dds_partition_prefix(self):
        random.seed(42)
        NumToTest = 1000
        names = set()
        for i in range(NumToTest):
            salobj.set_random_lsst_dds_partition_prefix()
            name = os.environ.get("LSST_DDS_PARTITION_PREFIX")
            self.assertTrue(name)
            names.add(name)
        # any duplicate names will reduce the size of names
        self.assertEqual(len(names), NumToTest)

    def test_modify_environ(self):
        rng = np.random.default_rng(seed=45)
        original_environ = os.environ.copy()
        n_to_delete = 3
        self.assertGreater(len(original_environ), n_to_delete)
        new_key0 = "_a_long_key_name_" + astropy.time.Time.now().isot
        new_key1 = "_another_long_key_name_" + astropy.time.Time.now().isot
        self.assertNotIn(new_key0, os.environ)
        self.assertNotIn(new_key1, os.environ)
        some_keys = rng.choice(list(original_environ.keys()), 3)
        kwargs = {
            some_keys[0]: None,
            some_keys[1]: None,
            some_keys[2]: "foo",
            new_key0: "bar",
            new_key1: None,
        }
        with salobj.modify_environ(**kwargs):
            for name, value in kwargs.items():
                if value is None:
                    self.assertNotIn(name, os.environ)
                else:
                    self.assertEqual(os.environ[name], value)
            for name, value in os.environ.items():
                if name in kwargs:
                    self.assertEqual(value, kwargs[name])
                else:
                    self.assertEqual(value, original_environ[name])
        self.assertEqual(os.environ, original_environ)

        # Values that are neither None nor a string should raise RuntimeError
        for bad_value in (3, 1.23, True, False):
            with self.assertRaises(RuntimeError):
                bad_kwargs = kwargs.copy()
                bad_kwargs[new_key1] = bad_value
                with salobj.modify_environ(**bad_kwargs):
                    pass
            self.assertEqual(os.environ, original_environ)

    async def test_domain_attr(self):
        async with salobj.Domain() as domain:
            self.assertEqual(domain.origin, os.getpid())

            self.assertEqual(domain.user_host, salobj.get_user_host())
            self.assertEqual(domain.default_identity, domain.user_host)
            self.assertEqual(domain.ackcmd_qos_set.profile_name, "AckcmdProfile")
            self.assertEqual(domain.command_qos_set.profile_name, "CommandProfile")
            self.assertEqual(domain.event_qos_set.profile_name, "EventProfile")
            self.assertEqual(domain.telemetry_qos_set.profile_name, "TelemetryProfile")
            self.assertTrue(domain.ackcmd_qos_set.volatile)
            self.assertTrue(domain.command_qos_set.volatile)
            self.assertFalse(domain.event_qos_set.volatile)
            self.assertTrue(domain.telemetry_qos_set.volatile)

    def test_index_generator(self):
        with self.assertRaises(ValueError):
            salobj.index_generator(1, 1)  # imin >= imax
        with self.assertRaises(ValueError):
            salobj.index_generator(1, 0)  # imin >= imax
        with self.assertRaises(ValueError):
            salobj.index_generator(0, 5, -1)  # i0 < imin
        with self.assertRaises(ValueError):
            salobj.index_generator(0, 5, 6)  # i0 > imax

        imin = -2
        imax = 5
        gen = salobj.index_generator(imin=imin, imax=imax)
        expected_values = [-2, -1, 0, 1, 2, 3, 4, 5, -2, -1, 0, 1, 2, 3, 4, 5, -2]
        values = [next(gen) for i in range(len(expected_values))]
        self.assertEqual(values, expected_values)

        imin = -2
        imax = 5
        i0 = 5
        expected_values = [5, -2, -1, 0, 1, 2, 3, 4, 5, -2]
        gen = salobj.index_generator(imin=imin, imax=imax, i0=i0)
        values = [next(gen) for i in range(len(expected_values))]
        self.assertEqual(values, expected_values)

    def test_name_to_name_index(self):
        for name, expected_result in (
            ("Script", ("Script", 0)),
            ("Script:0", ("Script", 0)),
            ("Script:15", ("Script", 15)),
            ("MTM1M3", ("MTM1M3", 0)),
            ("MTM1M3:47", ("MTM1M3", 47)),
        ):
            with self.subTest(name=name):
                result = salobj.name_to_name_index(name)
                self.assertEqual(result, expected_result)

        for bad_name in (
            (" Script:15"),  # leading space
            ("Script:15 "),  # trailing space
            ("Script:"),  # colon with no index
            ("Script:zero"),  # index is not an integer
        ):
            with self.subTest(bad_name=bad_name):
                with self.assertRaises(ValueError):
                    salobj.name_to_name_index(bad_name)

    def check_tai_from_utc(self, utc_ap):
        """Check tai_from_utc at a specific UTC date.

        Parameters
        ----------
        utc_ap : `astropy.time.Time`
            UTC date as an astropy time.
        """
        tai = salobj.tai_from_utc(utc_ap.utc.unix)

        tai_alt = alternate_tai_from_utc_unix(utc_ap.utc.unix)
        self.assertAlmostEqual(tai, tai_alt, delta=1e-6)

        tai2 = salobj.tai_from_utc(utc_ap.utc.iso, format="iso")
        self.assertAlmostEqual(tai, tai2, delta=1e-6)

        tai3 = salobj.tai_from_utc(utc_ap.utc.iso, format=None)
        self.assertAlmostEqual(tai, tai3, delta=1e-6)

        tai4 = salobj.tai_from_utc(utc_ap.utc.mjd, format="mjd")
        self.assertAlmostEqual(tai, tai4, delta=1e-6)

        tai_mjd = (tai + salobj.MJD_MINUS_UNIX_SECONDS) / salobj.SECONDS_PER_DAY
        tai_mjd_ap = astropy.time.Time(tai_mjd, scale="tai", format="mjd")
        tai5 = salobj.tai_from_utc(tai_mjd_ap)
        self.assertAlmostEqual(tai, tai5, delta=1e-6)

        tai_iso_ap = astropy.time.Time(utc_ap.tai.iso, scale="tai", format="iso")
        tai6 = salobj.tai_from_utc(tai_iso_ap)
        self.assertAlmostEqual(tai, tai6, delta=1e-6)

        tai7 = salobj.tai_from_utc(utc_ap)
        self.assertAlmostEqual(tai, tai7, delta=1e-6)

    def test_leap_second_table(self):
        """Check that the leap second table is set and an update scheduled."""
        self.assertIsNotNone(salobj.base._UTC_LEAP_SECOND_TABLE)
        update_timer = salobj.base._LEAP_SECOND_TABLE_UPDATE_TIMER
        self.assertTrue(update_timer.is_alive())
        self.assertTrue(update_timer.daemon)
        self.assertIs(update_timer.function, salobj.base._update_leap_second_table)
        update_margin = (
            salobj.base._LEAP_SECOND_TABLE_UPDATE_MARGIN_DAYS * salobj.SECONDS_PER_DAY
        )
        current_duration = (
            time.time() - salobj.base._UTC_LEAP_SECOND_TABLE[-1][0] - update_margin
        )
        self.assertGreater(update_timer.interval, current_duration)

    def test_tai_from_utc(self):
        """Test tai_from_utc."""
        # Check tai_from_utc near leap second transition at UTC = 2017-01-01
        # when leap seconds went from 36 to 37.
        utc0_ap = astropy.time.Time("2017-01-01", scale="utc", format="iso")
        for utc_ap in (
            (utc0_ap - 0.5 * u.day),
            (utc0_ap - 1 * u.second),
            (utc0_ap - 0.1 * u.second),
            (utc0_ap),
            (utc0_ap + 0.1 * u.second),
            (utc0_ap + 1 * u.second),
        ):
            with self.subTest(utc_ap=utc_ap):
                self.check_tai_from_utc(utc_ap=utc_ap)

        # Test values near the limits of the leap second table, which starts
        # on 1972-01-01T00:00:00 and ends whenever the table expires.
        first_utc_unix, first_tai_minus_utc = salobj.base._UTC_LEAP_SECOND_TABLE[0]
        desired_first_utc_unix = astropy.time.Time(
            "1972-01-01", scale="utc", format="iso"
        ).unix
        desired_first_tai_minus_utc = 10
        self.assertAlmostEqual(first_utc_unix, desired_first_utc_unix)
        self.assertAlmostEqual(first_tai_minus_utc, desired_first_tai_minus_utc)
        first_tai_unix = salobj.tai_from_utc(first_utc_unix)
        self.assertAlmostEqual(first_tai_unix, first_utc_unix + first_tai_minus_utc)
        with self.assertRaises(ValueError):
            salobj.tai_from_utc(first_utc_unix - 0.001)

        last_utc_unix, last_tai_minus_utc = salobj.base._UTC_LEAP_SECOND_TABLE[-1]
        # The last UTC that can be converted to TAI is a day earlier
        # than the last entry in the leap second table,
        # due to the possible 1 day smearing on the day before a leap second.
        last_usable_utc_unix = last_utc_unix - salobj.SECONDS_PER_DAY
        # last_utc_unix -= salobj.SECONDS_PER_DAY
        last_tai_unix = salobj.tai_from_utc(last_usable_utc_unix)
        # Final value of TAI-UTC in the table.
        # Note that the last entry in the table has TAI-UTC = None.
        final_tai_minus_utc = salobj.base._UTC_LEAP_SECOND_TABLE[-1][1]
        self.assertAlmostEqual(
            last_tai_unix - last_usable_utc_unix, final_tai_minus_utc
        )
        with self.assertRaises(ValueError):
            salobj.tai_from_utc(last_usable_utc_unix + 0.001)

    def test_utc_from_tai_unix(self):
        # Check utc_from_tai_unix near leap second transition at
        # UTC = 2017-01-01 when leap seconds went from 36 to 37.
        utc0 = astropy.time.Time("2017-01-01", scale="utc", format="iso").unix
        tai_minus_utc_before = 36
        tai_minus_utc_after = 37
        tai0 = utc0 + tai_minus_utc_after
        desired_tai_minus_utc_after = salobj.tai_from_utc_unix(utc0) - utc0
        self.assertAlmostEqual(tai_minus_utc_after, desired_tai_minus_utc_after)
        # Don't test right at tai0 because roundoff error could cause failure;
        # utc_from_tai_unix is discontinuous at a leap second tranition.
        for tai in (
            (tai0 - 0.5 * salobj.SECONDS_PER_DAY),
            (tai0 - 1),
            (tai0 - 0.001),
            (tai0 + 0.001),
            (tai0 + 1),
            (tai0 + 0.5 * salobj.SECONDS_PER_DAY),
        ):
            utc = salobj.utc_from_tai_unix(tai)
            tai_minus_utc = tai - utc
            if tai < tai0:
                self.assertAlmostEqual(tai_minus_utc, tai_minus_utc_before)
            else:
                self.assertAlmostEqual(tai_minus_utc, tai_minus_utc_after)

        # Test values near the limits of the TAI leap second table
        first_tai_unix, first_tai_minus_utc = salobj.base._TAI_LEAP_SECOND_TABLE[0]
        first_utc_unix = salobj.utc_from_tai_unix(first_tai_unix)
        desired_first_tai_minus_utc = 10
        self.assertAlmostEqual(first_tai_minus_utc, desired_first_tai_minus_utc)
        self.assertAlmostEqual(
            first_tai_unix - first_utc_unix, desired_first_tai_minus_utc
        )
        with self.assertRaises(ValueError):
            salobj.utc_from_tai_unix(first_tai_unix - 0.001)

        last_tai_unix, last_tai_minus_utc = salobj.base._TAI_LEAP_SECOND_TABLE[-1]
        last_utc_unix = salobj.utc_from_tai_unix(last_tai_unix)
        self.assertAlmostEqual(last_tai_unix - last_utc_unix, last_tai_minus_utc)
        with self.assertRaises(ValueError):
            salobj.utc_from_tai_unix(last_tai_unix + 0.001)

    def test_current_tai(self):
        utc0 = time.time()
        tai0 = salobj.tai_from_utc(utc0)
        # salobj.base.current_tai_from_utc uses tai_from_utc,
        # so it should give the same answer.
        tai1 = salobj.base.current_tai_from_utc()
        # salobj.current_tai uses the system TAI clock, if available.
        # This gives the correct answer (if your operating system
        # is correctly configured) and can differ from the answer given by
        # tai_from_utc by as much as a second on the day of a leap second.
        tai2 = salobj.current_tai()
        print(f"tai1-tai0={tai1-tai0:0.4f}")
        # The difference should be much less than 0.1
        # but pytest can introduce unexpected delays.
        self.assertLess(abs(tai1 - tai0), 0.1)
        self.assertGreaterEqual(tai1, tai0)
        # The difference between the value returned by current_tai
        # and current_tai_from_utc
        self.assertLess(abs(tai2 - tai0), 1.1)
        self.assertIs(type(tai0), float)
        self.assertIs(type(tai1), float)
        self.assertIs(type(tai2), float)

    def test_angle_diff(self):
        for angle1, angle2, expected_diff in (
            (5.15, 0, 5.15),
            (5.21, 359.20, 6.01),
            (270, -90, 0),
        ):
            with self.subTest(
                angle1=angle1, angle2=angle2, expected_diff=expected_diff
            ):
                diff = salobj.angle_diff(angle1, angle2)
                self.assertAlmostEqual(diff.deg, expected_diff)
                diff = salobj.angle_diff(angle2, angle1)
                self.assertAlmostEqual(diff.deg, -expected_diff)
                diff = salobj.angle_diff(Angle(angle1, u.deg), angle2)
                self.assertAlmostEqual(diff.deg, expected_diff)
                diff = salobj.angle_diff(angle1, Angle(angle2, u.deg))
                self.assertAlmostEqual(diff.deg, expected_diff)
                diff = salobj.angle_diff(Angle(angle1, u.deg), Angle(angle2, u.deg))
                self.assertAlmostEqual(diff.deg, expected_diff)

    def test_angle_wrap_center(self):
        for base_angle, expected_result in (
            (-180.001, 179.999),
            (-180, -180),
            (0, 0),
            (179.999, 179.999),
            (180, -180),
        ):
            for nwraps in (-2, -1, 0, 1, 2):
                with self.subTest(
                    base_angle=base_angle,
                    expected_result=expected_result,
                    nwraps=nwraps,
                ):
                    angle = base_angle + 360 * nwraps
                    result = salobj.angle_wrap_center(angle)
                    self.assertAlmostEqual(result.deg, expected_result)
                    result = salobj.angle_wrap_center(Angle(angle, u.deg))
                    self.assertAlmostEqual(result.deg, expected_result)

    def test_angle_wrap_nonnegative(self):
        for base_angle, expected_result in (
            (-0.001, 359.999),
            (0, 0),
            (180, 180),
            (359.999, 359.999),
            (360, 0),
        ):
            for nwraps in (-2, -1, 0, 1, 2):
                with self.subTest(
                    base_angle=base_angle,
                    expected_result=expected_result,
                    nwraps=nwraps,
                ):
                    angle = base_angle + 360 * nwraps
                    result = salobj.angle_wrap_nonnegative(angle)
                    self.assertAlmostEqual(result.deg, expected_result)
                    result = salobj.angle_wrap_nonnegative(Angle(angle, u.deg))
                    self.assertAlmostEqual(result.deg, expected_result)

    def test_assertAnglesAlmostEqual(self):
        for angle1, angle2 in ((5.15, 5.14), (-0.20, 359.81), (270, -90.1)):
            epsilon = Angle(1e-15, u.deg)
            with self.subTest(angle1=angle1, angle2=angle2):
                diff = abs(salobj.angle_diff(angle1, angle2))
                bad_diff = diff - epsilon
                self.assertGreater(bad_diff.deg, 0)
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(angle1, angle2, bad_diff)
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(angle1, angle2, bad_diff.deg)
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(angle2, angle1, bad_diff)
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(
                        Angle(angle1, u.deg), angle2, bad_diff
                    )
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(
                        angle1, Angle(angle2, u.deg), bad_diff
                    )
                with self.assertRaises(AssertionError):
                    salobj.assertAnglesAlmostEqual(
                        Angle(angle1, u.deg), Angle(angle2, u.deg), bad_diff
                    )

                good_diff = diff + epsilon
                salobj.assertAnglesAlmostEqual(angle1, angle2, good_diff)
                salobj.assertAnglesAlmostEqual(angle1, angle2, good_diff.deg)
                salobj.assertAnglesAlmostEqual(angle2, angle1, good_diff)
                salobj.assertAnglesAlmostEqual(Angle(angle1, u.deg), angle2, good_diff)
                salobj.assertAnglesAlmostEqual(angle1, Angle(angle2, u.deg), good_diff)
                salobj.assertAnglesAlmostEqual(
                    Angle(angle1, u.deg), Angle(angle2, u.deg), good_diff
                )


if __name__ == "__main__":
    unittest.main()
